# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Use an empty top-level docstring so Sphinx won't output the one below.
""""""

from textwrap import dedent as _dedent

"""NetKet error classes.
(Inspired by NetKet error classes)

=== When to create a NetKet error class?

If an error message requires more explanation than a one-liner, it is useful to
add it as a separate error class. This may lead to some duplication with
existing documentation or docstrings, but it will provide users with more help
when they are debugging a problem. We can also point to existing documentation
from the error docstring directly.

=== How to name the error class?

* If the error occurs when doing something, name the error
  <Verb><Object><TypeOfError>Error

  For instance, if you want to raise an error when applying a module with an
  invalid method, the error can be: ApplyModuleInvalidMethodError.

 <TypeOfError> is optional, for instance if there is only one error when
  modifying a variable, the error can simply be: ModifyVariableError.

* If there is no concrete action involved the only a description of the error is
  sufficient. For instance: InvalidFilterError, NameInUseError, etc.


=== Copy/pastable template for new error messages:

class Template(NetketError):
  "" "

  "" "
  def __init__(self):
    super().__init__(f'')
"""


class NetketError(Exception):
    def __init__(self, message):
        error_index = "https://netket.readthedocs.io/en/latest/api/errors.html"
        error_dir = "https://netket.readthedocs.io/en/latest/api/_generated/errors"
        module_name = self.__class__.__module__
        class_name = self.__class__.__name__
        error_msg = (
            f"{_dedent(message)}"
            f"\n"
            f"\n-------------------------------------------------------"
            f"\n"
            f"For more detailed informations, visit the following link:"
            f"\n\t {error_dir}/{module_name}.{class_name}.html"
            f"\n"
            f"or the list of all common errors at"
            f"\n\t {error_index}"
            f"\n-------------------------------------------------------"
            f"\n"
        )
        super().__init__(error_msg)


class NetketWarning(Warning):
    def __init__(self, message):
        error_index = "https://netket.readthedocs.io/en/latest/api/errors.html"
        error_dir = "https://netket.readthedocs.io/en/latest/api/_generated/errors"
        module_name = self.__class__.__module__
        class_name = self.__class__.__name__

        error_msg = (
            f"{_dedent(message)}"
            f"\n"
            f"\n-------------------------------------------------------"
            f"\n"
            f"For more detailed informations, visit the following link:"
            f"\n\t {error_dir}/{module_name}.{class_name}.html"
            f"\n"
            f"or the list of all common errors and warnings at"
            f"\n\t {error_index}"
            f"\n-------------------------------------------------------"
            f"\n"
        )
        super().__init__(error_msg)


#################################################
# Hilbert errors                                #
#################################################


class HilbertIndexingDuringTracingError(NetketError):
    """Illegal attempt to use state indexing function from
    inside of a Jax function transformation.

    This happens when you call functions such as
    :meth:`~netket.hilbert.DiscreteHilbert.states_to_numbers` or
    its opposite :meth:`~netket.hilbert.DiscreteHilbert.numbers_to_states` from
    within a :func:`jax.jit`, :func:`jax.grad`, :func:`jax.vjp` or similar jax
    function transformations.

    There is currently no workaround rather than returning the arrays from
    the jax function and performing the conversion outside of it.

    """

    def __init__(self):
        super().__init__(
            "\n"
            "Attempted to use state-indexing functions of an hilbert object "
            "inside a Jax function transformation (jax.jit, jax.grad & others)."
            "\n\n"
            "Functions that convert states to indices and vice-versa such as "
            "`hilbert.states_to_numbers()` or `hilbert.numbers_to_states()` "
            "are implemented in Numba and cannot be executed from within a"
            "jax function transformation."
        )


#################################################
# Operators errors                              #
#################################################


class NumbaOperatorGetConnDuringTracingError(NetketError):
    """Illegal attempt to use Numba-operators inside of a Jax
    function transformation.

    This happens when calling :meth:`~netket.operator.DiscreteOperator.get_conn_padded` or
    :meth:`~netket.operator.DiscreteOperator.get_conn_flattened` inside of a function
    that is being transformed by jax with transformations such as :func:`jax.jit`
    or :func:`jax.grad`, and the operator is not compatible with Jax.

    To avoid this error you can (i) convert your operator to a Jax compatible format if possible,
    or (ii) compute the connected elements outside of the jax function transformation and pass the
    results to a jax-transformed function.

    (i) Converting an Operator to a Jax compatible format
    -----------------------------------------------------

    Some operators can be converted to a jax-compatible format by calling the method
    `operator.to_jax_operator()`. If this method is not available or raises an error, it means
    that the operator cannot be converted.

    If the operator can be converted to a jax-compatible format, it will be possible to pass
    it as a standard argument to a jax-transformed function and it should not be declared as
    a static argument.

    Jax compatible operators can be used like standard operators, for example by passing
    it to :meth:`~netket.vqs.MCState.expect` function. However, the performance will differ from
    standard operators. In general, you might find that compile time will be much worse, while
    runtime might be faster or slower, depending on several factors.

    The biggest advantage to Jax operators, however, is when experimenting with jax code, as you
    can succesfully use them in your own custom functions as in the example below:

    .. code-block:: python

        import netket as nk
        import jax
        import jax.numpy as jnp

        graph = nk.graph.Chain(10)
        hilbert = nk.hilbert.Spin(1/2, graph.n_nodes)
        ham = nk.operator.Ising(hilbert, graph, h=1.0)
        ham_jax = ham.to_jax_operator()

        ma = nk.models.RBM()
        pars = ma.init(jax.random.PRNGKey(1), jnp.ones((2,graph.n_nodes)))

        samples = hilbert.all_states()

        @jax.jit
        def compute_local_energies(pars, ham_jax, s):
            # this would raise the error
            sp, mels = ham.get_conn_padded(s)
            # this will work
            sp, mels = ham_jax.get_conn_padded(s)

            logpsi_sp = ma.apply(pars, sp)
            logpsi_s = jnp.expand_dims(ma.apply(pars, s), -1)

            return jnp.sum(mels * jnp.exp(logpsi_sp-logpsi_s), axis=-1)

        elocs = compute_local_energies(pars, ham_jax, samples)
        elocs_grad = jax.jacrev(compute_local_energies)(pars, ham_jax, samples)

    .. note::

        Note that this transformation might be a relatively expensive operation, so you should avoid
        executing this inside of an hot loop.


    (ii) Precomputing connected elements outside of Jax transformations
    -------------------------------------------------------------------

    In most cases you won't be able to convert the operator to a Jax-compatible format.
    In those cases, the workaround we usually employ is to precompute the connected elements
    before entering the Jax context, splitting our function into a non-jitted function and
    into a jitted kernel.

    .. code-block:: python

      import netket as nk
      import jax
      import jax.numpy as jnp

      graph = nk.graph.Chain(10)
      hilbert = nk.hilbert.Spin(1/2, graph.n_nodes)
      ham = nk.operator.Ising(hilbert, graph, h=1.0)

      ma = nk.models.RBM()
      pars = ma.init(jax.random.PRNGKey(1), jnp.ones((2,graph.n_nodes)))

      samples = hilbert.all_states()

      def compute_local_energies(pars, ham, s):
          sp, mels = ham.get_conn_padded(s)
          return _compute_local_energies_kernel(pars, s, sp, mels)

      @jax.jit
      def _compute_local_energies_kernel(pars, s, sp, mels):
          logpsi_sp = ma.apply(pars, sp)
          logpsi_s = jnp.expand_dims(ma.apply(pars, s), -1)
          return jnp.sum(mels * jnp.exp(logpsi_sp-logpsi_s), axis=-1)

      elocs = compute_local_energies(pars, ham_jax, samples)
      elocs_grad = jax.jacrev(compute_local_energies)(pars, ham_jax, samples)


    Most :class:`~netket.operator.DiscreteOperator`s are implemented in Numba and therefore are not jax-compatible.
    To know if it's valid to use an operator inside of a jax function transformation you can
    check that it inherits from :class:`~netket.operator.DiscreteJaxOperator` by executing

    .. code-block:: python

        isinstance(operator, nk.operator.DiscreteJaxOperator)

    """

    def __init__(self, operator):
        operator_type = type(operator)
        super().__init__(
            f"\n"
            f"Attempted to use a Numba-based operator ({operator_type}) "
            f"inside a Jax function transformation (jax.jit, jax.grad & others)."
            f"\n\n"
            f"Numba-based operators are not compatible with Jax function "
            f"transformations, and can only be used outside of jax-function "
            f"boundaries."
            f"\n\n"
            f"Some operators can be converted to a Jax-compatible version by"
            f"calling `operator.to_jax_operator()`, but not all support it."
            f"If your operator does not support it, read the documentation to "
            f"find how to work-around this issue."
        )


class JaxOperatorSetupDuringTracingError(NetketError):
    """Illegal attempt to use a Jax-operator Numba-operators constructed inside of a Jax
    function transformation with non-constant data.

    This happens when building a :class:`~netket.operator.DiscreteJaxOperator` inside
    of a function that is being transformed by jax with transformations such as :func:`jax.jit`
    or :func:`jax.grad`, and the operator is not compatible with Jax.

    Notice that :class:`~netket.operator.DiscreteJaxOperator` can be used inside of jax
    function transformations, but NetKet is currently limited in that you cannot build
    them inside of Jax transformations.

    To avoid this error you should build your operators outside of the jax context.

    (i) Building a Jax operator outside of a jax context
    ----------------------------------------------------

    Build the operator outside of a jax context.

    .. code-block:: python

        import netket as nk
        import jax
        import jax.numpy as jnp

        N = 2

        ham = nk.operator.PauliStringsJax(['XI', 'IX'], jnp.array([0.3, 0.4]))

        samples = ham.hilbert.all_states()

        @jax.jit
        def compute_values(ham, s):
            return ham.get_conn_padded(s)

        compute_values(ham, samples)


    .. note::

        This limitation is not systematic, and it could be lifted in the future by some
        interested coder. If at that moment Jax will support dynamic shape, this feature could
        be implemented at no additional runtime cost. If Jax won't support yet dynamic shapes
        then it should be implemented as a secondary path (instead of this error) that is only
        taken if the operator is constructed inside of a jax context. This path will lead to a
        slightly less optimized, higher computational cost operators.

        If you are really interested in contributing to NetKet and find yourself in need of
        building operators in a jax context (for example because you're doing optimal control),
        get in touch with us by opening an issue.

    .. note::

        Most operators lazily initialise the fields used to compute the connected elements only
        when needed. To check whether an operator was initialized you can probe the boolean flag
        :code:`operator._initialized`. If :code:`operator._initialized` is True, you can safely call
        :meth:`~netket.operator.DiscreteJaxOperator.get_conn_padded` and similar methods. If it is
        False, then the setup procedure will be handled by an internal method usually called
        :code:`operator._setup()`. If you see this error, it means that this method internally
        uses dynamically determined shapes, and it is what should be converted to be jax-friendly.


    """

    def __init__(self, operator):
        operator_type = type(operator)
        super().__init__(
            f"\n"
            f"Attempted to initialise a jax operator ({operator_type}) "
            f"inside a Jax function transformation (jax.jit, jax.grad & others)."
            f"\n\n"
            f"Jax-based operators cannot be initialised (constructed) within Jax "
            f"function transformations, and can only be initialised outside Jax "
            f"function boundaries."
            f"\n\n"
            f"To avoid this error, construct the operator outside of the jax"
            f"function and pass it to it as a standard argument."
        )


class JaxOperatorNotConvertibleToNumba(NetketError):
    """Illegal attempt to convert to the Numba format a Jax operator that had been flattened
    and unflattened.

    This probably happened because you passed a Jax operator to a jax function transformation
    or jitted function and then tried to re-convert it to the numba format like in the example
    below:

    .. code-block:: python

        import netket as nk

        hi = nk.hilbert.Spin(0.5, 2)

        op = nk.operator.spin.sigmax(hi, 0)
        op = op.to_jax_operator()

        @jax.jit
        def test(op):
            op.to_numba_operator()

        test(op)

    Unfortunately, once an operator is flattened with {ref}`jax.tree_util.tree_flatten`, which
    happens at all jax-function transformation boundaries, it usually cannot be converted back to
    the original numba form.

    This happens for performance reasons, and we might reconsider. If it is a problem for you, do
    open an issue.

    """

    def __init__(self, operator):
        super().__init__(
            "\n"
            "Illegal attempt to convert to the Numba format a Jax operator that had been flattened "
            "and unflattened."
            "\n\n"
            "Jax-based operators cannot be flattened or passed to a jax function and then be "
            "converted to their numba format."
        )


#################################################
# Jacobian and QGT errors                       #
#################################################


class IllegalHolomorphicDeclarationForRealParametersError(NetketError):
    """Illegal attempt to declare a function holomorphic when it has some
    or all real parameters, which makes it automatically non-holomorphic.

    This error may arise when computing the Jacobian directly or when
    constructing the Quantum Geometric Tensor, which internally constructs
    the Jacobian of your variational function.

    In general we could silence this error automatically and ignore the
    :code:`holomorphic=True` argument, but we wish to stress with the users
    the importance of correctly declaring this argument.

    To solve this error, simply stop declaring :code:`holomorphic=True`, by
    either removing it or specifying :code:`holomorphic=False`.
    """

    def __init__(self):
        super().__init__(
            """
        A function with real parameters is not holomorphic.

        You declared `holomorphic=True` when computing the Jacobian, Quantum
        Geometric Tensor or a similar, but your variational function has
        real parameters, so it cannot be holomorphic.

        To fix this error, remove the keyword argument `holomorphic=True`.
        """
        )


class NonHolomorphicQGTOnTheFlyDenseRepresentationError(NetketError):
    """
    QGTOnTheFly cannot be converted to a dense matrix for non-holomorphic
    functions which have complex parameters.

    This limitation does not apply if the parameters are all real.

    This error might have happened for two reasons:
     - you specified `holomorphic=False` because your ansatz is non-holomorphic.
       In that case you should use `QGTJacobianPyTree` or `QGTJacobianDense`
       implementations.

     - you did not specify `holomorphic`, which leads to the default value of
       `holomorphic=False` (in that case, you should have seen a warning). If
       that is the case, you should carefully check if your ansatz is holomorhic
       almost everywhere, and if that is the case, specify `holomorphic=True`.
       If your ansatz is not-holomorphic, the same suggestion as above applies.

    Be warned that if you specify `holomorphic=True` when your ansatz is mathematically
    not holomorphic is a surprisingly bad idea and will lead to numerically wrong results,
    so I'd invite you not to lie to your computer.
    """

    def __init__(self):
        super().__init__(
            """
            QGTOnTheFly cannot be converted to a dense matrix for non-holomorphic
            functions which have complex parameters.

            This limitation does not apply if the parameters are all real.

            This error might have happened for two reasons:
             - you specified `holomorphic=False` because your ansatz is non-holomorphic.
               In that case you should use `QGTJacobianPyTree` or `QGTJacobianDense`
               implementations.

             - you did not specify `holomorphic`, which leads to the default value of
               `holomorphic=False` (in that case, you should have seen a warning). If
               that is the case, you should carefully check if your ansatz is holomorhic
               almost everywhere, and if that is the case, specify `holomorphic=True`.
               If your ansatz is not-holomorphic, the same suggestion as above applies.

            Be warned that if you specify `holomorphic=True` when your ansatz is mathematically
            not holomorphic is a surprisingly bad idea and will lead to numerically wrong results,
            so I'd invite you not to lie to your computer.
            """
        )


class HolomorphicUndeclaredWarning(NetketWarning):
    """
    Complex-to-Complex model detected. Defaulting to :code:`holomorphic = False`
    for the calculation of its jacobian.

    However, :code:`holomorphic = False` might lead to slightly increased
    computational cost, some disabled features and/or worse quality of
    solutions found with iterative solvers.
    If your model is actually holomorphic, you should specify :code:`holomorphic = True`
    to unblock some extra, possibly more performant algorithms.

    If you are unsure whether your variational function is holomorphic or not,
    you should check if it satisfies the
    `Cauchy-Riemann equations <https://en.wikipedia.org/wiki/Cauchy–Riemann_equations>`_.

    To check numerically those conditions on a random set of samples you can use the
    function :func:`netket.utils.is_probably_holomorphic`. If this function returns
    False then your ansatz is surely not holomorphic, while if it returns True your
    ansatz is likely but not guaranteed to be holomorphic.

    To check those conditions numerically, you can check by following this
    example:

    .. code:: python

        hi = nk.hilbert.Spin(0.5, 2)
        sa = nk.sampler.MetropolisLocal(hi)
        ma = nk.models.RBM(param_dtype=complex)
        # construct the variational state
        vs = nk.vqs.MCState(sa, ma)

        nk.utils.is_probably_holomorphic(vs._apply_fun,
                                         vs.parameters,
                                         vs.samples,
                                         model_state = vs.model_state)

    To suppress this warning correctly specify the keyword argument
    `holomorphic`.

    .. note::
        a  detailed discussion, explaining how to easily check those conditions
        analitically is found in the documentation of
        :func:`netket.utils.is_probably_holomorphic`).

    """

    def __init__(self):
        super().__init__(
            """
            Defaulting to `holomorphic=False`, but this might lead to increased
            computational cost or disabled features. Check if your variational
            function is holomorphic, and if so specify `holomorphic=True`as an extra
            keyword argument.

            To silence this warning, specify the `holomorphic=False/True` keyword
            argument.

            To numerically check whether your variational function is or not holomorphic
            you can use the following snippet:

            ```python
               vs = nk.vqs.MCState(...)

               nk.utils.is_probably_holomorphic(vs._apply_fun, vs.parameters, vs.samples, vs.model_state)
            ```

            if `nk.utils.is_probably_holomorphic` returns False, then your function is not holomorphic.
            If it returns True, it is probably holomorphic.
            """
        )


class RealQGTComplexDomainError(NetketError):
    """
    This error is raised when you apply the Quantum Geometric Tensor of a
    non-holomorphic function to a complex-valued vector, because the
    operation is not well-defined.

    As is explained in the documentation of the geomtric tensors, the QGT
    implementation for non-holomorphic functions corresponds to the real-
    part of the QGT, not the full QGT.

    This is because in most applications of variational monte carlo, such as
    the Time-Dependent Variational Principle, Ground-state search or
    supervised-learning, you only require knowledge of the real part, and
    can safely discard the imaginary part that would incur in an increased
    computational cost.

    While the product of the real part of QGT by a complex vector is well-defined,
    to prevent the common mistake of assuming that the QGT is complex we
    explicitly raise this error, forcing users to manually multiply the QGT
    by the real and imaginary part of the vector, as we would have to do
    inside of this class anyway.

    If this is really the mathematical operation you want to perform, then
    you can do it manually, but very often we have found that when you
    apply the real part of the QGT to a complex vector you might have
    your math wrong. In such cases, sometimes what you actually wanted
    to do was

    .. code:: python

       >>> import netket as nk; import jax
       >>>
       >>> vstate = nk.vqs.FullSumState(nk.hilbert.Spin(0.5, 5), \
                                        nk.models.RBM(param_dtype=complex))
       >>> _, vec = vstate.expect_and_grad(nk.operator.spin.sigmax(vstate.hilbert, 1))
       >>> G = nk.optimizer.qgt.QGTOnTheFly(vstate, holomorphic=False)
       >>>
       >>> vec_real = jax.tree.map(lambda x: x.real, vec)
       >>> sol = G@vec_real

   Or, if you used the QGT in a linear solver, try using:

   .. code:: python

       >>> import netket as nk; import jax
       >>>
       >>> vstate = nk.vqs.FullSumState(nk.hilbert.Spin(0.5, 5), \
                                        nk.models.RBM(param_dtype=complex))
       >>> _, vec = vstate.expect_and_grad(nk.operator.spin.sigmax(vstate.hilbert, 1))
       >>>
       >>> G = nk.optimizer.qgt.QGTOnTheFly(vstate, holomorphic=False)
       >>> vec_real = jax.tree.map(lambda x: x.real, vec)
       >>>
       >>> linear_solver = jax.scipy.sparse.linalg.cg
       >>> solution, info = G.solve(linear_solver, vec_real)

    """

    def __init__(self):
        super().__init__(
            """
            Cannot multiply the (real part of the) QGT by a complex vector.
            You should either take the real part of the vector, or perform
            the multiplication against the real and imaginary part of the
            vector separately and then recomposing the two.

            This is happening because you have real parameters or a non-holomorphic
            complex wave function. In this case, the Quantum Geometric Tensor object
            only stores the real part of the QGT.

            If you were executing a matmul `G@vec`, try using:

            .. code:: python

               >>> vec_real = jax.tree_util.tree_map(lambda x: x.real, vec)
               >>> G@vec_real

            If you used the QGT in a linear solver, try using:

            .. code:: python

               >>> vec_real = jax.tree_util.tree_map(lambda x: x.real, vec)
               >>> G.solve(linear_solver, vec_real)

            to fix this error.

            Be careful whether you need the real or imaginary part
            of the vector in your equations!
            """
        )


class SymmModuleInvalidInputShape(NetketError):
    """
    This error when you attempt to use a module with an input having a wrong number
    of dimensions.

    In particular, Simmetric layers require inputs with 3 dimensions :math:`(B, C, L)`:
        - Batch dimension, which should be 1 if only 1 sample is considered
        - Channel or Features dimension, which should encode multiple features, usually
            originated from previous layers. If this is the first simmetric layer, you
            can set this dimension to 1
        - Length, which should span the physical degrees of freedom.
    """

    def __init__(self, name, x):
        super().__init__(
            f"""
            Input to DenseSymmFFT has {x.ndim =} but 3 are required.

            The input format is (B,C,L), aka (batches, features/channels, length). If
            you have a single sample, simply use `jnp.atleast_3d(x)` before calling
            this module.

            If this is the first layer in a network, you usually need to set the channel
            dimension to 1.
            """
        )


class UnoptimalSRtWarning(NetketWarning):
    """
    SRt should be used when the number of parameters exceed the number of samples.
    If this is not the case, employing `netket.driver.VMC` with the `nk.optimizer.SR` preconditioner
    is a more efficient option while maintaining the same parameter dynamics.

    .. note::
        a  detailed discussion can be found in the documentation of
        :func:`netket.experimental.driver.VMC_SRt`).

    """

    def __init__(self, n_parameters, n_samples):
        super().__init__(
            f"""
            You are in the case n_samples > num_params ({n_samples} > {n_parameters}),
            for which the `VMC_SRt` is not optimal. Consider using `netket.driver.VMC`
            with the preconditioner `nk.optimizer.SR` to achieve the same parameter dynamics,
            but with improved speed.
            """
        )


class InvalidConstraintInterface(NetketError):
    """Error thrown when a constraint for an Homogeneous
    Hilbert space does not conform to the expected interface.

    A custom Constraint must respect the :class:`netket.hilbert.constraint.DiscreteHilbertConstraint`
    interface by inheriting that class and by having a jax-jittable :code:`__call__` method.

    If you see this error, your constraint might just be a function without inheriting from this
    class, or it does not have a jittable call method.

    See the class documentation of :class:`netket.hilbert.constraint.DiscreteHilbertConstraint` for
    some examples.
    """

    def __init__(self):
        super().__init__(
            """Your custom constraint does not respect the constraint interface, probably because
            it does not inherit from `netket.hilbert.constraint.DiscreteHilbertConstraint`, because it does not
            have a jittable `__call__` method.

            Look at the documentation online to learn more.
            """
        )


class UnhashableConstraintError(NetketError):
    """
    Error thrown when a constraint is not a fully static pytree, and it cannot be hashed.

    This error is thrown when a constraint is a PyTree with some leaf nodes.
    This usually happens because you did not mark all the pytree fields in your constraint
    as `struct.field(pytree_node=False)`.

    For a good example, see the documentation of :class:`netket.hilbert.constraint.DiscreteHilbertConstraint`.
    Below, you find a coincise example of how to fix this error:

    .. code-block:: python

        from typing import Any
        from netket.utils import struct
        from netket.hilbert.index.constraints import DiscreteHilbertConstraint

        class MyCustomConstraint(DiscreteHilbertConstraint):
            fieldA: Any = struct.field(pytree_node=False)
            fieldB: Any = struct.field(pytree_node=False)
            def __init__(self, fieldA, fieldB):
                self.fieldA = fieldA
                self.fieldB = fieldB

            def __call__(self, x):
                ....

            def __hash__(self):
                return hash(("MyCustomConstraint", self.fieldA, self.fieldB))

            def __eq__(self, other):
                if isinstance(other, UnhashableConstraintError):
                    return self.fieldA == other.fieldA and self.fieldB == other.fieldB
                return False

    """

    def __init__(self, constraint):
        constraintT = type(constraint)
        super().__init__(
            f"""
            The custom constraint type {constraintT} is a pytree with some nodes, meaning that it cannot be
            safely passed to jax as a static argument.

            This probably happened because some of the fields in your constraint are not marked as
            `nk.utils.struct.field(pytree_node=False)`.

            To fix this error, you should mark all fields in your constraint as per the example below:
                from typing import Any
                from netket.utils import struct

                class MyCustomConstraint(nk.hilbert.constraint.DiscreteHilbertConstraint):
                    fieldA: Any = struct.field(pytree_node=False)
                    fieldB: Any = struct.field(pytree_node=False)
                    def __init__(self, fieldA, fieldB):
                        self.fieldA = fieldA
                        self.fieldB = fieldB

                    def __call__(self, x):
                        ....

                    def __hash__(self):
                        return hash(("MyCustomConstraint", self.fieldA, self.fieldB))

                    def __eq__(self, other):
                        if isinstance(other, UnhashableConstraintError):
                            return self.fieldA == other.fieldA and self.fieldB == other.fieldB
                        return False

            """
        )


class UnoptimisedCustomConstraintRandomStateMethodWarning(NetketWarning):
    """
    Warning thrown when calling `random_state` on a Hilbert space with a custom constraint.

    This warning is thrown when the custom Hilbert space constraint does not have a custom
    `random_state` method implemented. This will default to a slow, possibly infinitely-looping
    method to generate random states.

    The default fallback works by generating random states in the unconstrained Hilbert space
    until one is found that satisfies the constraint. This can be very slow, and even never
    terminate if the constraint is too restrictive.

    .. note::

        Generating random states is only required when initializing a Markov Chain Monte Carlo
        sampler, but is generally not needed during on. Therefore, even if this warning is thrown,
        it might not be a problem if you do not reset your chains very often.

        In general, if your constraint is not overly (exponentially) restrictive, you may not
        need to worry about this warning.

    .. note::

        The fallback implementation is not optimised, especially on GPUs. It will generate 1 random
        state at a time until it finds one that satisfies the constraint, and will repeat this for
        every different chain.

        We welcome contributions to improve this method to perform the loop in batches.

    .. note::

        You can silence this warning by setting the environment variable ``NETKET_RANDOM_STATE_FALLBACK_WARNING=0``
        or by setting ``nk.config.netket_random_state_fallback_warning = 0`` in your code.


    Example implementation of a custom `random_state` method:
    ---------------------------------------------------------

    Implementations of :func:`netket.hilbert.random_state` are dispatched based on the Hilbert space and
    constraint type using Plum's multiple dispatch (see the link <https://beartype.github.io/plum/intro.html>_

    See an example here

    .. code-block:: python

        @dispatch.dispatch
        def random_state(hilb: HilbertType, constraint: ConstraintType, key, batches: int, *, dtype=None):
            # your custom implementation here
            # You should return a batch of `batches` random states, with the given dtype.
            # return jax.Array with shape (batches, hilb.size) and dtype dtype.


    .. note::

        We welcome contributions to improve the documentation here.

    """

    def __init__(self, hilbert_space, constraint):
        hilb_typ = f"{type(hilbert_space).__module__}.{type(hilbert_space).__name__}"
        constraint_typ = f"{type(constraint).__module__}.{type(constraint).__name__}"
        super().__init__(
            f"""
            Defaulting to a slow, possibly infinitely-looping method to generate random state of
            the current Hilbert space with a custom constraint. Consider implementing a
            custom `random_state` method for your constraint if this method takes a long time to
            generate a random state.

            ================================================================
            You can silence this warning by setting the environment variable
            ``NETKET_RANDOM_STATE_FALLBACK_WARNING=0``
            or by setting ``nk.config.netket_random_state_fallback_warning = False``
            in your code.
            ================================================================

            To generate a custom random_state dispatched method, you should use multiple dispatch
            following the following syntax:

            >>> import netket as nk
            >>> from netket.utils import dispatch
            >>>
            >>> @dispatch.dispatch
            >>> def random_state(hilb: {hilb_typ},
                                constraint: {constraint_typ},
                                key,
                                batches: int,
                                *,
                                dtype=None):
            >>>    # your custom implementation here
            >>>    # You should return a batch of `batches` random states, with the given dtype.
            >>>    # return jax.Array with shape (batches, hilb.size) and dtype dtype.

            """
        )


class NetKetPyTreeUndeclaredAttributeAssignmentError(AttributeError, NetketError):
    """
    Error thrown when trying to assign an undeclared attribute to a NetKet-style Pytree.

    This error is thrown when you try to assign an attribute to a NetKet-style Pytree ( a class
    inheriting from :class:`netket.utils.struct.Pytree`) that was not declared in the class definition.

    This error is thrown to prevent you from accidentally creating a new attribute that is not
    part of the Pytree structure, which would lead to unexpected behaviour.

    To fix this error, you should declare the attribute in the class definition, as shown in the example below:

    .. code-block:: python

        from netket.utils import struct

        class MyPytree(struct.Pytree):
            # This line below was probably missing in your class definition
            my_attribute: Any

            def __init__(self, my_attribute):
                self.my_attribute = my_attribute

    Note that if the field is not a jax-Array or another Pytree, you should instead declare it as a static
    or non-node field, which will trigger recompilation when passed to jax functions, as shown below:

    .. code-block:: python

        from netket.utils import struct
        import jax

        class MyPytree(struct.Pytree):
            my_dynamic_attribute: jax.Array
            my_static_attribute : int = struct.field(pytree_node=False)

            def __init__(self, dyn_val, static_val):
                self.my_dynamic_attribute = dyn_val
                self.my_static_attribute = static_val

        leafs, structure = jax.tree.flatten(MyPytree(1, 2))
        print(leafs)
        # [1]
        print(structure)
        # PyTreeDef(CustomNode(MyPytree[{'_pytree__node_fields': ('my_dynamic_attribute',), 'my_static_attribute': 2}], [*]))

    From which you should see that the list of leafs contains only the dynamic attribute, while the
    structure, which is static information, holds the static attribute.

    .. note::

        You can also declare the class to have dynamic nodes, in which case the attributes will be inferred
        automatically during the class initialization. However, this is not recomended as it is impossible
        to declare nodes that are not jax-Arrays or Pytrees.

        Regardless, if you wish to do so you should declare the class as follows:

        .. code-block:: python

            from netket.utils import struct

            class MyPytree(struct.Pytree, dynamic_nodes=True):
                def __init__(self, my_attribute):
                    self.my_attribute = my_attribute

    """

    def __init__(self, pytree, attribute, valid_attributes):
        pytreeT = f"{type(pytree).__module__}.{type(pytree).__name__}"
        super().__init__(
            f"""
            Tried to assign an undeclared attribute "{attribute}" to the Pytree class "{pytreeT}" during
            the __init__ method.

            Declared attributes for "{pytreeT}" are "{valid_attributes}".

            This error is thrown when you try to assign an attribute to a NetKet-style Pytree ( a class that
            inherits from `nk.utils.struct.Pytree`) that was not declared in the class definition.

            To fix this error, simply declare the attributes in the class definition as shown in the example below:

                from netket.utils import struct
                import jax

                class MyPytree(struct.Pytree):
                    my_dynamic_attribute: jax.Array
                    my_static_attribute : int = struct.field(pytree_node=False)

                    def __init__(self, dyn_val, static_val):
                        self.my_dynamic_attribute = dyn_val
                        self.my_static_attribute = static_val

            """
        )


class UndeclaredSpinOderingWarning(NetketWarning):
    """
    Warning thrown when a Spin Hilbert space is created without specifying the spin ordering.

    This warning is thrown in the transition period of NetKet 3.14 to NetKet 3.15 (september to
    december 2024), to warn users that the correspondence between ±1 and spin up/down will change
    according to the table below.

    +----------+----------------------------------+-----------------------------------+
    | State    | Old Behavior                     | New Behavior                      |
    |          | :code:`inverted_ordering=True`   | :code:`inverted_ordering=False`   |
    +==========+==================================+===================================+
    | ↑ ↑ ↑    | -1 -1 -1                         | +1 +1 +1                          |
    +----------+----------------------------------+-----------------------------------+
    | ↑ ↑ ↓    | -1 -1 +1                         | +1 +1 -1                          |
    +----------+----------------------------------+-----------------------------------+
    | ↑ ↓ ↑    | -1 +1 -1                         | +1 -1 +1                          |
    +----------+----------------------------------+-----------------------------------+
    | ↑ ↓ ↓    | -1 +1 +1                         | +1 -1 -1                          |
    +----------+----------------------------------+-----------------------------------+
    | ↓ ↑ ↑    | +1 -1 -1                         | -1 +1 +1                          |
    +----------+----------------------------------+-----------------------------------+
    | ↓ ↑ ↓    | +1 -1 +1                         | -1 +1 -1                          |
    +----------+----------------------------------+-----------------------------------+
    | ↓ ↓ ↑    | +1 +1 -1                         | -1 -1 +1                          |
    +----------+----------------------------------+-----------------------------------+
    | ↓ ↓ ↓    | +1 +1 +1                         | -1 -1 -1                          |
    +----------+----------------------------------+-----------------------------------+

    The old behaviour is the default behaviour of NetKet 3.14 and before, while the new
    behaviour will become the default starting 1st january 2025.
    For that reason, in the transition period, we will print warnings asking to explicitly
    specify which ordering you want

    .. warning::

        The ordering of the Spin Hilbert space basis has historically always been
        such that `-1=↑, 1=↓`, but it will be changed 1st january 2025 to
        be such that `1=↑, -1=↓`.

        The change will break:
            - code that relies on the assumption that -1=↑;
            - all saves because the inputs to the network will change;
            - custom operators that rely on the basis being ordered;

        To avoid distruption, NetKet will support **both** conventions in the (near)
        future. You can specify the ordering you need with :code:`inverted_ordering = True`
        (historical ordering) or :code:`inverted_ordering=False` (future default behaviour).

        If you do not specify this flag, a future version of NetKet might break your
        serialized weights or other logic, so we strongly reccomend that you either
        limit yourself to NetKet 3.14, or that you specify :code:`inverted_ordering`
        explicitly.

    To avoid this warning, you can :

     - explicitly specify the ordering you want wit the `inverted_ordering` flag, which will
       ensure that your code will still work in the future with no changes. If possible,
       we suggest you use the new ordering `inverted_ordering=False`, but if you want to
       keep loading parameters serialized before th switch, you should use `inverted_ordering=True`.

     - You can silence this warning by setting the environment variable
       ``NETKET_SPIN_ORDERING_WARNING=0`` or by setting
       ``nk.config.netket_spin_ordering_warning = False`` in your code.

    """

    def __init__(self):
        super().__init__(
            _dedent(
                """
                You have not explicitly specified the spin ordering for the Hilbert space.
                The default behaviour is currently `-1=↑, 1=↓`, but it will be changed 1st january 2025 to `1=↑, -1=↓`.

                - To maintain the current behaviour in the future, specify `inverted_ordering=True` (this
                    allows you to load NN parameters you have saved in the past)
                - To opt-in today in the future default, specify `inverted_ordering=False` (so your code will
                    work without changes in the future)

                If you do not care about this warning, you can silence it by setting the environment variable
                `NETKET_SPIN_ORDERING_WARNING=0` or by executing `nk.config.netket_spin_ordering_warning = False`

                This warning will be shown once per day during interactive sessions, and always in scripts and MPI/SLURM jobs unless silenced.
                """
            )
        )


class ParameterMismatchError(NetketError):
    """
    Error thrown when the structure of the parameters does not match the expected structure.

    This error is thrown when you attempt to modify the `parameters` or `variables` attribute of a
    :class:`~netket.vqs.VariationalState` with a structure that does not match the previous structure.

    """

    def __init__(self, expected_structure, structure):
        super().__init__(
            f"""
            The structure of the parameters does not match the expected structure.

            Expected structure: {expected_structure}
            Structure of the parameters: {structure}

            This error is because you attempted to modify the ``parameters`` or ``variables`` attribute of a
            variational state with a structure that does not match the previous structure.

            To fix this error, you should ensure that the structure of the parameters you are trying to assign
            matches the structure of the parameters that were already present in the variational state.

            If you believe this error was thrown in error, or it prevents you from doing something, please open an issue.
            """
        )


#################################################
# Functions to throw errors                     #
#################################################


def concrete_or_error(force, value, error_class, *args, **kwargs):
    """
    Wraps `jax.core.concrete_or_error` but allows us to throw our
    own errors.

    Args:
      force: function to be executed on the value (for example np.asarray)
        that would fail if value is not concrete
      value: the argument to `force`
      error: the constructor of the custom error to throw.
      *args: any additional argument and keyword argument to pass to the custom
        error type constructor.
    """

    from jax.core import ConcretizationTypeError, concrete_or_error

    try:
        return concrete_or_error(
            force,
            value,
            """
          """,
        )
    except ConcretizationTypeError as err:
        raise error_class(*args, **kwargs) from err


#################################################
# Graph errors                                  #
#################################################


class InitializePeriodicLatticeOnSmallLatticeWarning(NetketWarning):
    """
    Warning thrown when attempting to create a periodic lattice on a lattice with less than two sites in one direction.

    In a periodic lattice with two sites (a, b) in one direction,
    the expected behavior is to have an edge connecting both a->b and b->a

    However, as the lattice uses an undirected graph to represent the lattice, it does not support this behavior.
    Hence, only one edge a->b is created, which makes the behavior equivalent to an open boundary condition in this direction.

    This may cause unexpected behavior if you intend to loop over the edges to create a Hamiltonian, for example.

    To avoid this warning, consider either using a lattice with more than two sites in the direction you want to be periodic,
    or define the graph using :class:`~netket.graph.Graph` by adding the edges manually.
    """

    def __init__(self, extent, dimension):
        super().__init__(
            f"""
            You are attempting to define a lattice with length {extent} in dimension {dimension} using periodic boundary condition.

            Lattice with less than two sites in one direction does not support periodic boundary condition.
            The behavior of the lattice is equivalent to an open boundary condition in this direction.

            To avoid this warning, consider either using a lattice with more than two sites in the direction you want to be periodic,
            or define the graph using :class:`~netket.graph.Graph` by adding the edges manually.
            """
        )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from netket.utils.types import DType

from netket.graph import AbstractGraph
from netket.hilbert import AbstractHilbert

from ._local_operator import LocalOperator


def check_acting_on_subspace(acting_on_subspace, hilbert, graph):
    """Check `acting_on_subspace` argument used by various operators."""
    if acting_on_subspace is None:
        acting_on_subspace = list(range(hilbert.size))
    elif isinstance(acting_on_subspace, int):
        start = acting_on_subspace
        acting_on_subspace = [start + i for i in range(graph.n_nodes)]
    elif isinstance(acting_on_subspace, list):
        if len(acting_on_subspace) != graph.n_nodes:
            raise ValueError(
                "acting_on_subspace must be a list of length graph.n_nodes"
            )
    else:
        raise TypeError("acting_on_subspace must be a list or single integer")

    return acting_on_subspace


class GraphOperator(LocalOperator):
    r"""
    A graph-based quantum operator. In its simplest terms, this is the sum of
    local operators living on the edge of an arbitrary graph.
    """

    def __init__(
        self,
        hilbert: AbstractHilbert,
        graph: AbstractGraph,
        site_ops=[],
        bond_ops=[],
        bond_ops_colors=[],
        dtype: DType | None = None,
        *,
        acting_on_subspace: None | list[int] | int = None,
    ):
        r"""
        A graph-based quantum operator. In its simplest terms, this is the sum of
        local operators living on the edge of an arbitrary graph.

        A ``GraphOperator`` is constructed giving a hilbert space and either a
        list of operators acting on sites or a list acting on the bonds.
        Users can specify the color of the bond that an operator acts on, if
        desired. If none are specified, the bond operators act on all edges.

        Args:
         hilbert: Hilbert space the operator acts on.
         graph: The graph whose vertices and edges are considered to construct the
                operator
         site_ops: A list of operators in matrix form that act
                on the nodes of the graph.
                The default is an empty list. Note that if no site_ops are
                specified, the user must give a list of bond operators.
         bond_ops: A list of operators that act on the edges of the graph.
             The default is None. Note that if no bond_ops are
             specified, the user must give a list of site operators.
         bond_ops_colors: A list of edge colors, specifying the color each
             bond operator acts on. The default is an empty list.
         dtype: Data type of the matrix elements.
         acting_on_subspace: Specifies the mapping between nodes of the graph and
            Hilbert space sites, so that graph node :code:`i ∈ [0, ..., graph.n_nodes]`,
            corresponds to :code:`acting_on_subspace[i] ∈ [0, ..., hilbert.n_sites]`.
            Must be a list of length `graph.n_nodes`. Passing a single integer :code:`start`
            is equivalent to :code:`[start, ..., start + graph.n_nodes - 1]`.

        Examples:
         Constructs a ``GraphOperator`` operator for a 2D system.

         >>> import netket as nk
         >>> sigmax = [[0, 1], [1, 0]]
         >>> mszsz = [[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]]
         >>> edges = [[0, 1], [1, 2], [2, 3], [3, 4], [4, 5], [5, 6], [6, 7], [7, 8],
         ... [8, 9], [9, 10], [10, 11], [11, 12], [12, 13], [13, 14], [14, 15],
         ... [15, 16], [16, 17], [17, 18], [18, 19], [19, 0]]
         >>> g = nk.graph.Graph(edges=edges)
         >>> hi = nk.hilbert.Spin(0.5, N=g.n_nodes)
         >>> op = nk.operator.GraphOperator(
         ... hi, site_ops=[sigmax], bond_ops=[mszsz], graph=g)
         >>> print(op)
         GraphOperator(dim=20, #acting_on=40 locations, constant=0.0, dtype=float64, graph=Graph(n_nodes=20, n_edges=20))
        """
        acting_on_subspace = check_acting_on_subspace(
            acting_on_subspace, hilbert, graph
        )
        self._acting_on_subspace = acting_on_subspace

        # Ensure that at least one of SiteOps and BondOps was initialized
        if len(bond_ops) == 0 and len(site_ops) == 0:
            raise ValueError("Must input at least site_ops or bond_ops.")

        # Create the local operator as the sum of all site and bond operators
        operators = []
        acting_on = []

        # Site operators
        if len(site_ops) > 0:
            for i in range(graph.n_nodes):
                for site_op in site_ops:
                    i_prime = acting_on_subspace[i]
                    operators.append(site_op)
                    acting_on.append([i_prime])

        # Bond operators
        if len(bond_ops_colors) > 0:
            if len(bond_ops) != len(bond_ops_colors):
                raise ValueError(
                    """The GraphHamiltonian definition is inconsistent.
                    The sizes of bond_ops and bond_ops_colors do not match."""
                )

            if len(bond_ops) > 0:
                #  Use edge_colors to populate operators
                for u, v, color in graph.edges(return_color=True):
                    u, v = acting_on_subspace[u], acting_on_subspace[v]
                    for c, bond_color in enumerate(bond_ops_colors):
                        if bond_color == color:
                            operators.append(bond_ops[c])
                            acting_on.append([u, v])
        else:
            assert len(bond_ops) == 1

            for u, v in graph.edges():
                u, v = acting_on_subspace[u], acting_on_subspace[v]
                operators.append(bond_ops[0])
                acting_on.append([u, v])

        super().__init__(hilbert, operators, acting_on, dtype=dtype)
        self._graph = graph

    @property
    def graph(self) -> AbstractGraph:
        """The graph on which this Operator is defined"""
        return self._graph

    @property
    def acting_on_subspace(self):
        """
        Mapping between nodes of the graph and Hilbert space sites as given in
        the constructor.
        """
        return self._acting_on_subspace

    def copy(self, *, dtype: DType | None = None):
        """Returns a copy of the operator, while optionally changing the dtype
        of the operator.

        Args:
            dtype: optional dtype
        """
        return super().copy(dtype=dtype, _cls=LocalOperator)

    def __repr__(self):
        ao = self.acting_on

        acting_str = f"acting_on={ao}"
        if len(acting_str) > 55:
            acting_str = f"#acting_on={len(ao)} locations"
        return f"{type(self).__name__}(dim={self.hilbert.size}, {acting_str}, constant={self.constant}, dtype={self.dtype}, graph={self.graph})"



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import numpy as np

import jax
from jax import numpy as jnp
from flax import linen as nn
from jax.nn.initializers import zeros, normal

from netket.utils.types import NNInitFunc
from netket import jax as nkjax
from netket import nn as nknn

default_kernel_init = normal(stddev=0.001)


class PureRBM(nn.Module):
    """
    Encodes the pure state |ψ><ψ|, because it acts on row and column
    indices with the same RBM.
    """

    param_dtype: Any = np.float64
    """The dtype of the weights."""
    activation: Any = nknn.log_cosh
    """The nonlinear activation function."""
    alpha: float | int = 1
    """feature density. Number of features equal to alpha * input.shape[-1]"""
    use_hidden_bias: bool = True
    """if True uses a bias in the dense layer (hidden layer bias)."""
    use_visible_bias: bool = True
    """if True adds a bias to the input not passed through the nonlinear layer."""
    precision: Any = None
    """Numerical precision of the computation see :class:`jax.lax.Precision` for details."""

    kernel_init: NNInitFunc = default_kernel_init
    """Initializer for the Dense layer matrix."""
    hidden_bias_init: NNInitFunc = zeros
    """Initializer for the hidden bias."""
    visible_bias_init: NNInitFunc = zeros
    """Initializer for the visible bias."""

    @nn.compact
    def __call__(self, σr, σc, symmetric=True):
        W = nn.Dense(
            name="Dense",
            features=int(self.alpha * σr.shape[-1]),
            param_dtype=self.param_dtype,
            use_bias=self.use_hidden_bias,
            kernel_init=self.kernel_init,
            bias_init=self.hidden_bias_init,
            precision=self.precision,
        )
        xr = self.activation(W(σr)).sum(axis=-1)
        xc = self.activation(W(σc)).sum(axis=-1)

        if symmetric:
            y = xr + xc
        else:
            y = xr - xc

        if self.use_visible_bias:
            v_bias = self.param(
                "visible_bias",
                self.visible_bias_init,
                (σr.shape[-1],),
                self.param_dtype,
            )
            if symmetric:
                out_bias = jnp.dot(σr + σc, v_bias)
            else:
                out_bias = jnp.dot(σr - σc, v_bias)

            y = y + out_bias

        return 0.5 * y


class MixedRBM(nn.Module):
    """
    Encodes the pure state |ψ><ψ|, because it acts on row and column
    indices with the same RBM.
    """

    param_dtype: Any = np.float64
    """The dtype of the weights."""
    activation: Any = nknn.log_cosh
    """The nonlinear activation function."""
    alpha: float | int = 1
    """feature density. Number of features equal to alpha * input.shape[-1]"""
    use_bias: bool = True
    """if True uses a bias in the dense layer (hidden layer bias)."""
    precision: Any = None
    """numerical precision of the computation see :class:`jax.lax.Precision` for details."""

    kernel_init: NNInitFunc = default_kernel_init
    """Initializer for the Dense layer matrix."""
    bias_init: NNInitFunc = zeros
    """Initializer for the hidden bias."""

    @nn.compact
    def __call__(self, σr, σc):
        U_S = nn.Dense(
            name="Symm",
            features=int(self.alpha * σr.shape[-1]),
            param_dtype=self.param_dtype,
            use_bias=False,
            kernel_init=self.kernel_init,
            precision=self.precision,
        )
        U_A = nn.Dense(
            name="ASymm",
            features=int(self.alpha * σr.shape[-1]),
            param_dtype=self.param_dtype,
            use_bias=False,
            kernel_init=self.kernel_init,
            precision=self.precision,
        )
        y = U_S(0.5 * (σr + σc)) + 1j * U_A(0.5 * (σr - σc))

        if self.use_bias:
            bias = self.param(
                "bias",
                self.bias_init,
                (int(self.alpha * σr.shape[-1]),),
                nkjax.dtype_real(self.param_dtype),
            )
            y = y + bias

        y = self.activation(y)
        return y.sum(axis=-1)


class NDM(nn.Module):
    """
    Encodes a Positive-Definite Neural Density Matrix using the ansatz from Torlai and
    Melko, PRL 120, 240503 (2018).

    Assumes real dtype.
    A discussion on the effect of the feature density for the pure and mixed part is
    given in Vicentini et Al, PRL 122, 250503 (2019).
    """

    param_dtype: Any = np.float64
    """The dtype of the weights."""
    activation: Any = nknn.log_cosh
    """The nonlinear activation function."""
    alpha: float | int = 1
    """The feature density for the pure-part of the ansatz.
    Number of features equal to alpha * input.shape[-1]
    """
    beta: float | int = 1
    """The feature density for the mixed-part of the ansatz.
    Number of features equal to beta * input.shape[-1]
    """
    use_hidden_bias: bool = True
    """if True uses a bias in the dense layer (hidden layer bias)."""
    use_ancilla_bias: bool = True
    """if True uses a bias in the dense layer (hidden layer bias)."""
    use_visible_bias: bool = True
    """if True adds a bias to the input not passed through the nonlinear layer."""
    precision: Any = None
    """numerical precision of the computation see :class:`jax.lax.Precision` for details."""

    kernel_init: NNInitFunc = default_kernel_init
    """Initializer for the Dense layer matrix."""
    bias_init: NNInitFunc = default_kernel_init
    """Initializer for the hidden bias."""
    visible_bias_init: NNInitFunc = default_kernel_init
    """Initializer for the visible bias."""

    @nn.compact
    def __call__(self, σ):
        σr, σc = jax.numpy.split(σ, 2, axis=-1)

        ψ_S = PureRBM(
            name="PureSymm",
            alpha=self.alpha,
            activation=self.activation,
            param_dtype=self.param_dtype,
            use_hidden_bias=self.use_hidden_bias,
            use_visible_bias=self.use_visible_bias,
            visible_bias_init=self.visible_bias_init,
            kernel_init=self.kernel_init,
            hidden_bias_init=self.bias_init,
            precision=self.precision,
        )

        ψ_A = PureRBM(
            name="PureASymm",
            alpha=self.alpha,
            activation=self.activation,
            param_dtype=self.param_dtype,
            use_hidden_bias=self.use_hidden_bias,
            use_visible_bias=self.use_visible_bias,
            visible_bias_init=self.visible_bias_init,
            kernel_init=self.kernel_init,
            hidden_bias_init=self.bias_init,
            precision=self.precision,
        )

        Π = MixedRBM(
            name="Mixed",
            alpha=self.beta,
            param_dtype=self.param_dtype,
            use_bias=self.use_ancilla_bias,
            activation=self.activation,
            kernel_init=self.kernel_init,
            bias_init=self.bias_init,
            precision=self.precision,
        )
        return (
            ψ_S(σr, σc, symmetric=True) + 1j * ψ_A(σr, σc, symmetric=False) + Π(σr, σc)
        )



# Copyright 2021-2024 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import tensor_networks

from .rbm import RBM, RBMModPhase, RBMMultiVal, RBMSymm
from .equivariant import GCNN
from .full_space import LogStateVector
from .jastrow import Jastrow
from .gaussian import Gaussian
from .deepset import DeepSetRelDistance, DeepSetMLP
from .ndm import NDM
from .autoreg import (
    AbstractARNN,
    ARNNSequential,
    ARNNDense,
    ARNNConv1D,
    ARNNConv2D,
)
from .fast_autoreg import (
    FastARNNSequential,
    FastARNNDense,
    FastARNNConv1D,
    FastARNNConv2D,
)
from .mlp import MLP

from .slater import Slater2nd, MultiSlater2nd

from .utils import update_GCNN_parity


from netket.utils import _hide_submodules

_hide_submodules(__name__)



# Copyright 2024 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from flax import linen as nn
import jax
from jax import numpy as jnp
from jax.nn.initializers import normal

from netket.hilbert import HomogeneousHilbert
from netket.utils.types import NNInitFunc
from netket.jax import dtype_complex
from netket.utils.types import DType

default_kernel_init = normal(stddev=0.01)


class MPSPeriodic(nn.Module):
    r"""
    A periodic Matrix Product State (MPS) for a quantum state of discrete
    degrees of freedom.
    The MPS is defined as

    .. math:: \Psi(s_1,\dots s_N) = \mathrm{Tr} \left[ A[s_1] \dots A[s_N] \right] ,

    for arbitrary local quantum numbers :math:`s_i`, where :math:`A[s_i]` are matrices of shape :math:`(\text{bond_dim}, \text{bond_dim})` for  :math:`i=1, \dots N` , depending on the value of the local quantum number :math:`s_i`.
    """

    hilbert: HomogeneousHilbert
    """Hilbert space on which the state is defined."""
    bond_dim: int
    """Bond dimension of the MPS tensors. See formula above."""
    symperiod: int | None = None
    """
    Periodicity in the chain of MPS tensors.
    The chain of MPS tensors is constructed as a sequence of identical
    unit cells consisting of symperiod tensors. if None, symperiod equals the
    number of physical degrees of freedom.
    """
    unroll: int = 1
    """the number of scan iterations to unroll within a single iteration of a loop."""
    checkpoint: bool = True
    """Whether to use jax.checkpoint on the scan function for memory efficiency."""
    kernel_init: NNInitFunc = default_kernel_init
    """the initializer for the MPS weights. This is added to an identity tensor."""
    param_dtype: DType = float
    """complex or float, whether the variational parameters of the MPS are real or complex."""

    def setup(self):
        L, d, D = self.hilbert.size, self.hilbert.local_size, self.bond_dim
        self._L, self._d, self._D = L, d, D

        # determine the shape of the unit cell
        if self.symperiod is None:
            self._symperiod = L
        else:
            self._symperiod = self.symperiod

        if L % self._symperiod == 0 and self._symperiod > 0:
            unit_cell_shape = (self._symperiod, d, D, D)
        else:
            raise AssertionError(
                "The number of degrees of freedom of the Hilbert space needs to be a multiple of the period of the MPS"
            )

        # define diagonal tensors with correct unit cell shape
        iden_tensors = jnp.repeat(
            jnp.eye(D, dtype=self.param_dtype)[jnp.newaxis, :, :],
            self._symperiod * d,
            axis=0,
        )
        iden_tensors = iden_tensors.reshape(self._symperiod, d, D, D)

        self.tensors = (
            self.param("tensors", self.kernel_init, unit_cell_shape, self.param_dtype)
            + iden_tensors
        )

    def __call__(self, x):
        """
        Queries the tensor network for the input configurations x.

        Args:
            x: the input configuration
        """
        x = jnp.atleast_1d(x)
        assert x.shape[-1] == self.hilbert.size

        x_shape = x.shape
        if jnp.ndim(x) != 2:
            x = jax.lax.collapse(x, 0, -1)

        qn_x = self.hilbert.states_to_local_indices(x)

        ψ = jax.vmap(self.contract_mps, in_axes=0)(qn_x)
        ψ = ψ.reshape(x_shape[:-1])

        return jnp.log(ψ.astype(dtype_complex(self.param_dtype)))

    def contract_mps(self, qn):
        """
        Internal function, used to contract the tensor network with some input
        tensor.

        Args:
            qn: The input tensor to be contracted with this MPS
        """
        # create all tensors in mps from unit cell
        all_tensors = jnp.tile(self.tensors, (self._L // self._symperiod, 1, 1, 1))

        edge = jnp.eye(self._D, dtype=self.param_dtype)

        def base_scan_func(edge, pair):
            tensor, qn = pair
            edge = edge @ tensor[qn, :]
            return edge, None

        # Apply jax.checkpoint conditionally to the base_scan_func
        scan_func = (
            jax.checkpoint(base_scan_func) if self.checkpoint else base_scan_func
        )

        edge, _ = jax.lax.scan(scan_func, edge, (all_tensors, qn), unroll=self.unroll)

        return jnp.trace(edge)


class MPSOpen(nn.Module):
    r"""
    An open Matrix Product State (MPS) for a quantum state of discrete
    degrees of freedom.
    The MPS is defined as

    .. math:: \Psi(s_1,\dots s_N) = \mathrm{Tr} \left[ A[s_1]\dots A[s_N] \right] ,

    for arbitrary local quantum numbers :math:`s_i`, where :math:`A[s_{i}]` are vectors of shape :math:`(\text{bond_dim},)` for :math:`i=1,N` and matrices
    of shape :math:`(\text{bond_dim}, \text{bond_dim})` for  :math:`i=2, \dots N-1` , depending on the value of the local quantum number :math:`s_i`.
    """

    hilbert: HomogeneousHilbert
    """Hilbert space on which the state is defined."""
    bond_dim: int
    """Bond dimension of the MPS tensors. See formula above."""
    unroll: int = 1
    """the number of scan iterations to unroll within a single iteration of a loop."""
    checkpoint: bool = True
    """Whether to use jax.checkpoint on the scan function for memory efficiency."""
    kernel_init: NNInitFunc = default_kernel_init
    """the initializer for the MPS weights. This is added to an identity tensor."""
    param_dtype: DType = float
    """complex or float, whether the variational parameters of the MPS are real or complex."""

    def setup(self):
        L, d, D = self.hilbert.size, self.hilbert.local_size, self.bond_dim
        self._L, self._d, self._D = L, d, D
        self.param_dtype_cplx = dtype_complex(self.param_dtype)

        iden_boundary_tensor = jnp.ones((d, D), dtype=self.param_dtype)
        self.left_tensors = (
            self.param("left_tensors", self.kernel_init, (d, D), self.param_dtype)
            + iden_boundary_tensor
        )
        self.right_tensors = (
            self.param("right_tensors", self.kernel_init, (d, D), self.param_dtype)
            + iden_boundary_tensor
        )

        # determine the shape of the unit cell
        unit_cell_shape = (L - 2, d, D, D)

        iden_tensors = jnp.repeat(
            jnp.eye(D, dtype=self.param_dtype)[jnp.newaxis, :, :],
            (L - 2) * d,
            axis=0,
        )
        iden_tensors = iden_tensors.reshape(L - 2, d, D, D)
        self.middle_tensors = (
            self.param(
                "middle_tensors", self.kernel_init, unit_cell_shape, self.param_dtype
            )
            + iden_tensors
        )

    def __call__(self, x):
        """
        Queries the tensor network for the input configurations x.

        Args:
            x: the input configuration
        """
        x = jnp.atleast_1d(x)
        assert x.shape[-1] == self.hilbert.size
        x_shape = x.shape
        if x.ndim != 2:
            x = jax.lax.collapse(x, 0, -1)
        qn_x = self.hilbert.states_to_local_indices(x)

        ψ = jax.vmap(self.contract_mps)(qn_x)
        ψ = ψ.reshape(x_shape[:-1])

        return jnp.log(ψ.astype(dtype_complex(self.param_dtype)))

    def contract_mps(self, qn):
        """
        Internal function, used to contract the tensor network with some input
        tensor.

        Args:
            qn: The input tensor to be contracted with this MPS
        """

        edge = self.left_tensors[qn[0], :]

        def base_scan_func(edge, pair):
            tensor, qn = pair
            edge = edge @ tensor[qn, :]
            return edge, None

        # Apply jax.checkpoint conditionally to the base_scan_func
        scan_func = (
            jax.checkpoint(base_scan_func) if self.checkpoint else base_scan_func
        )

        edge, _ = jax.lax.scan(
            scan_func, edge, (self.middle_tensors, qn[1:-1]), unroll=self.unroll
        )
        psi = self.right_tensors[qn[-1], :] @ edge
        return psi



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import jax
from jax import numpy as jnp

import numpy as np

from netket import jax as nkjax
from netket.experimental.hilbert import ContinuousHilbert, Particle
from netket.utils.dispatch import dispatch


def take_sub(key, x, n):
    key, subkey = jax.random.split(key)
    ind = jax.random.choice(
        subkey, jnp.arange(0, x.shape[0], 1), replace=False, shape=(n,)
    )
    return x[ind, :]


@dispatch
def random_state(hilb: Particle, key, batches: int, *, dtype):
    """If no periodic boundary conditions are present particles are positioned normally distributed around the origin.

    If periodic boundary conditions are present the particles are positioned uniformly inside the box and a small
    gaussian noise is added on top.
    If periodic boundary conditions are chosen only for certain dimensions, the periodic initialization is used for
    all of those dimensions and the free initialization is used for all the other ones.
    """
    pbc = np.array(hilb.n_particles * hilb.geometry.pbc)
    boundary = np.tile(pbc, (batches, 1))

    # TODO genearalize to particles with diferent domains
    Ls = np.array(hilb.n_particles * hilb.domain)

    # If we have PBCs this defines the size of the Gaussian noise in the particle positions.
    # Without PBCs this is never needed (initialization with random Gaussian)
    # and we can set a random number (here 1.)
    modulus = np.where(np.equal(pbc, False), 1.0, Ls)

    min_modulus = np.min(modulus)

    # use real dtypes because this does not work with complex ones.
    gaussian = jax.random.normal(
        key, shape=(batches, hilb.size), dtype=nkjax.dtype_real(dtype)
    )

    width = min_modulus / (4.0 * hilb.n_particles)
    # The width gives the noise level. In the periodic case the
    # particles are evenly distributed between 0 and min(L). The
    # distance between the particles coordinates is therefore given by
    # min(L) / hilb.N. To avoid particles to have coincident
    # positions the noise level should be smaller than half this distance.
    # We choose width = min(L) / (4*hilb.N)
    noise = gaussian * width

    key = jax.random.split(key, num=batches)
    sdim = len(hilb.domain)
    n = int(np.ceil(hilb.n_particles ** (1 / sdim)))
    xs = jnp.linspace(0, min(hilb.domain), n)
    uniform = jnp.array(jnp.meshgrid(*(sdim * [xs]))).reshape(-1, sdim)
    uniform = jnp.tile(uniform, (batches, 1, 1))

    uniform = jax.vmap(take_sub, in_axes=(0, 0, None))(
        key, uniform, hilb.n_particles
    ).reshape(batches, -1)
    rs = jnp.where(np.equal(boundary, False), gaussian, (uniform + noise) % modulus)

    return jnp.asarray(rs, dtype=dtype)


@dispatch
def flip_state_scalar(hilb: ContinuousHilbert, key, x, i):
    raise TypeError(
        "Flipping state is undefined for continuous Hilbert spaces. "
        "(Maybe you tried using `MetropolisLocal` on a continuous Hilbert space? "
        "This won't work because 'flipping' a continuous variable is not defined. "
        "You should try a different sampler.)"
    )



# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
import jax.numpy as jnp
import numpy as np
from functools import partial

from flax import struct

from .base import MetropolisRule
import netket.jax as nkjax


class LangevinRule(MetropolisRule):
    r"""
    A transition rule that uses Langevin dynamics [1] to update samples.

    .. math::
       x_{t+dt} = x_t + dt \nabla_x \log p(x) \vert_{x=x_t} + \sqrt{2 dt}\eta,

    where  :math:`\eta` is normal distributed noise :math:`\eta \sim \mathcal{N}(0,1)`.
    This rule only works for continuous Hilbert spaces.

    [1]: https://en.wikipedia.org/wiki/Metropolis-adjusted_Langevin_algorithm
    """

    dt: float
    """Time step in the Langevin dynamics."""
    chunk_size: int | None = struct.field(pytree_node=False)
    """Chunk size for computing gradients of the ansatz."""

    def __init__(self, dt: float = 0.001, chunk_size: int | None = None):
        """
        Constructs the Langevin Hastings proposal rule.

        Args:
            dt: the timestep for the langevin dynamics (default=0.001)
            chunk_size: optional chunk size to reduce memory usage
        """
        self.dt = dt
        self.chunk_size = chunk_size

    def transition(rule, sampler, machine, parameters, state, key, r):
        if jnp.issubdtype(r.dtype, jnp.complexfloating):
            raise TypeError("LangevinRule does not work with complex basis elements.")

        n_chains = r.shape[0]
        hilb = sampler.hilbert

        # TODO generalize to different types of geometry per particle
        pbc = np.array(hilb.n_particles * hilb.geometry.pbc, dtype=r.dtype)
        boundary = np.tile(pbc, (n_chains, 1))

        Ls = np.array(hilb.n_particles * hilb.domain, dtype=r.dtype)
        modulus = np.where(np.equal(pbc, False), jnp.inf, Ls)

        # one langevin step
        rp, log_corr = _langevin_step(
            key,
            r,
            machine.apply,
            parameters,
            sampler.machine_pow,
            rule.dt,
            chunk_size=rule.chunk_size,
            return_log_corr=True,
        )

        rp = jnp.where(np.equal(boundary, False), rp, rp % modulus)

        return rp, log_corr

    def __repr__(self):
        return f"LangevinRule(dt={self.dt})"


@partial(jax.jit, static_argnames=("apply_fun", "chunk_size", "return_log_corr"))
def _langevin_step(
    key,
    r,
    apply_fun,
    parameters,
    machine_pow,
    dt,
    chunk_size=None,
    return_log_corr=True,
):
    """Single step of samples with Langevin dynamics"""

    n_chains, hilb_size = r.shape

    # steps with Langevin dynamics
    noise_vec = jax.random.normal(key, shape=(n_chains, hilb_size), dtype=r.dtype)

    def _log_prob(x):
        """Conversion to a log probability"""
        return machine_pow * apply_fun(parameters, x).real

    def _single_grad(x):
        """Derivative of log_prob with respect to a single sample x"""
        x = x.reshape(x.shape[-1])
        g = nkjax.grad(lambda xi: _log_prob(xi).ravel()[0])(x)
        return g if jnp.iscomplexobj(r) else g.real

    grad_logp_r = nkjax.vmap_chunked(_single_grad, chunk_size=chunk_size)(r)

    rp = r + dt * grad_logp_r + jnp.sqrt(2 * dt) * noise_vec

    if not return_log_corr:
        return rp
    else:
        log_q_xp = -0.5 * jnp.sum(noise_vec**2, axis=-1)
        grad_logp_rp = nkjax.vmap_chunked(_single_grad, chunk_size=chunk_size)(rp)
        log_q_x = -jnp.sum((r - rp - dt * grad_logp_rp) ** 2, axis=-1) / (4 * dt)

        return rp, log_q_x - log_q_xp



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netket.utils.mpi.mpi import mpi4py_available, mpi4jax_available
from netket.utils.mpi import rank, n_nodes

from ._cpu_info import available_cpus


def check_mpi():
    """
    When called via::

        # python3 -m netket.tools.check_mpi
        mpi4puavailable              : True
        mpi4jax_available            : True
        available_cpus (rank 0)       : 12
        n_nodes                      : 1
        mpi4py | MPI version         : (3, 1)
        mpi4py | MPI library_version : Open MPI v4.1.0, ...

    this will print out basic MPI information to allow users to check whether
    the environment has been set up correctly.
    """
    if rank > 0:
        return

    info = {
        "mpi4py_available": mpi4py_available,
        "mpi4jax_available": mpi4jax_available,
        "available_cpus (rank 0)": available_cpus(),
    }
    if mpi4py_available:
        from mpi4py import MPI

        info.update(
            {
                "n_nodes": n_nodes,
                "mpi4py | MPI version": MPI.Get_version(),
                "mpi4py | MPI library_version": MPI.Get_library_version(),
            }
        )

    maxkeylen = max(len(k) for k in info.keys())

    for k, v in info.items():
        print(f"{k:{maxkeylen}} : {v}")


check_mpi()



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .rnn import RNN, LSTMNet, GRUNet1D
from .fast_rnn import FastRNN, FastLSTMNet, FastGRUNet1D

from netket.models import Slater2nd as _deprecated_Slater2nd
from netket.models import MultiSlater2nd as _deprecated_MultiSlater2nd

_deprecations = {
    # May 2024
    "Slater2nd": (
        "netket.experimental.models.Slater2nd is deprecated: use "
        "netket.models.Slater2nd (netket >= 3.12)",
        _deprecated_Slater2nd,
    ),
    "MultiSlater2nd": (
        "netket.experimental.models.MultiSlater2nd is deprecated: use "
        "netket.models.MultiSlater2nd (netket >= 3.12)",
        _deprecated_MultiSlater2nd,
    ),
}

from netket.utils.deprecation import deprecation_getattr as _deprecation_getattr

__getattr__ = _deprecation_getattr(__name__, _deprecations)
del _deprecation_getattr



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netket.utils.dispatch import dispatch

from netket.operator import (
    DiscreteOperator,
    AbstractSuperOperator,
    Squared,
    DiscreteJaxOperator,
)

from netket.vqs.mc import (
    kernels,
    check_hilbert,
    get_local_kernel_arguments,
    get_local_kernel,
)

from .state import MCMixedState


@dispatch
def get_local_kernel_arguments(vstate: MCMixedState, Ô: DiscreteOperator):  # noqa: F811
    check_hilbert(vstate.diagonal.hilbert, Ô.hilbert)

    σ = vstate.diagonal.samples
    σp, mels = Ô.get_conn_padded(σ)
    return σ, (σp, mels)


@dispatch
def get_local_kernel_arguments(  # noqa: F811
    vstate: MCMixedState, Ô: DiscreteJaxOperator
):
    check_hilbert(vstate.diagonal.hilbert, Ô.hilbert)

    σ = vstate.diagonal.samples
    σp, mels = Ô.get_conn_padded(σ)
    return σ, (σp, mels)


@dispatch
def get_local_kernel_arguments(  # noqa: F811
    vstate: MCMixedState, Ô: AbstractSuperOperator
):  # noqa: F811
    check_hilbert(vstate.hilbert, Ô.hilbert)
    σ = vstate.samples
    σp, mels = Ô.get_conn_padded(σ)
    return σ, (σp, mels)


@dispatch
def get_local_kernel(vstate: MCMixedState, Ô: AbstractSuperOperator):  # noqa: F811
    return kernels.local_value_kernel


@dispatch
def get_local_kernel(vstate: MCMixedState, Ô: DiscreteOperator):  # noqa: F811
    return kernels.local_value_op_op_cost


@dispatch
def get_local_kernel(vstate: MCMixedState, Ô: DiscreteJaxOperator):  # noqa: F811
    return kernels.local_value_op_op_cost


@dispatch
def get_local_kernel_arguments(  # noqa: F811
    vstate: MCMixedState, Ô: Squared[AbstractSuperOperator]
):
    check_hilbert(vstate.hilbert, Ô.hilbert)

    σ = vstate.samples
    σp, mels = Ô.parent.get_conn_padded(σ)
    return σ, (σp, mels)


@dispatch
def get_local_kernel(  # noqa: F811
    vstate: MCMixedState, Ô: Squared[AbstractSuperOperator]
):
    return kernels.local_value_squared_kernel



import dataclasses
import typing as tp
from typing import Any
from collections.abc import Callable
from abc import ABCMeta
from copy import copy
from functools import partial
from types import MappingProxyType

import jax
import jax.numpy as jnp

from flax import serialization

from .fields import CachedProperty, _cache_name, _raw_cache_name, Uninitialized
from netket.utils import config
from netket.errors import NetKetPyTreeUndeclaredAttributeAssignmentError

P = tp.TypeVar("P", bound="Pytree")

DATACLASS_USER_INIT_N_ARGS = "_pytree_n_args_max"
"""
variable name used by dataclasses inheriting from a pytree to
store the topmost non-dataclass class in a mro.
"""


def identity(x):
    return x


class PytreeMeta(ABCMeta):
    """
    Metaclass for PyTrees, takes care of initializing and turning
    frozen PyTrees to immutable after __init__.
    """

    def __call__(cls: type[P], *args: tp.Any, **kwargs: tp.Any) -> P:
        __tracebackhide__ = True

        obj: P = cls.__new__(cls, *args, **kwargs)

        # set things before init
        vars_dict = vars(obj)

        # Set 'default' values
        for field in obj._pytree__cachedprop_fields:
            vars_dict[field] = Uninitialized

        for field, default_fn in cls._pytree__default_setters.items():
            if field not in vars_dict:
                vars_dict[field] = default_fn()

        obj.__dict__["_pytree__initializing"] = True
        try:
            obj.__init__(*args, **kwargs)
        finally:
            del obj.__dict__["_pytree__initializing"]

        # set things after init
        if obj._pytree__class_dynamic_nodes:
            vars_dict["_pytree__node_fields"] = tuple(
                sorted(
                    field
                    for field in vars_dict
                    if field not in cls._pytree__static_fields
                )
            )
        else:
            vars_dict["_pytree__node_fields"] = cls._pytree__data_fields

        return obj


class Pytree(metaclass=PytreeMeta):
    """
    Astract Base class for jaw-aware PyTree classes.

    A class inheriting from PyTree can be passed to a jax function
    as a standard argument, and can contain both static and dynamic
    fields. Those will be correctly handled when flattening the
    PyTree.

    Static fields must be specified as class attributes, by specifying
    the :func:`netket.utils.struct.field` as
    :code:`struct.field(pytree_node=False)`.

    Example:
        Construct a PyTree with a 'constant' value

        >>> from netket.utils.struct import field, Pytree
        >>> import jax
        >>>
        >>> class MyPyTree(Pytree):
        ...     a: int = field(pytree_node=False)
        ...     b: jax.Array
        ...     def __init__(self, a, b):
        ...         self.a = a
        ...         self.b = b
        ...     def __repr__(self):
        ...         return f"MyPyTree(a={self.a}, b={self.b})"
        >>>
        >>> my_pytree = MyPyTree(1, jax.numpy.ones(2))
        >>> jax.jit(lambda x: print(x))(my_pytree)  # doctest:+ELLIPSIS
            MyPyTree(a=1, b=Traced...


    PyTree classes by default are not mutable, therefore they
    behave similarly to frozen dataclasses. If you want to
    make a PyTree mutable, you can specify the `mutable=True`
    argument in the class definition.

    Example:

        >>> from netket.utils.struct import field, Pytree
        >>>
        >>> class MyPyTree(Pytree, mutable=True):
        ...     a: int = field(pytree_node=False)
        ...     ...

    By default only the fields declared as class attributes
    can be set and/or modified after initialization. If you
    want to allow the creation of new fields after initialization,
    you can specify the `dynamic_nodes=True` argument in the
    class definition.

    PyTree classes can also be inherited by a netket dataclass,
    in which case the dataclass will be initialized with the
    fields of the PyTree. However, this behaviour is deprecated
    and will be removed in the future. We suggest you to remove
    the `@nk.utils.struct.dataclass` decorator and simply define
    an `__init__` method.
    """

    _pytree__initializing: bool
    _pytree__class_is_mutable: bool
    _pytree__fields: dict[str, dataclasses.Field]
    _pytree__static_fields: tuple[str, ...]
    _pytree__node_fields: tuple[str, ...]
    _pytree__ignored_fields: tuple[str, ...]
    _pytree__setter_descriptors: frozenset[str]

    _pytree__cachedprop_fields: tuple[str, ...]

    def __init_subclass__(cls, mutable: bool = None, dynamic_nodes: bool = False):
        super().__init_subclass__()

        # gather class info
        class_vars = vars(cls)
        setter_descriptors = set()
        all_fields, data_fields, static_fields, ignored_fields = _inherited_fields(cls)

        # new
        default_setters = _inherited_defaults_setters(cls)  # _pytree__default_setters
        cached_prop_fields = set()

        # Custom information
        noserialize_fields = _inherited_noserialize_fields(cls)
        sharded_fields = _inherited_sharded_fields(cls)

        # add special static fields that are in the instance, not in the class
        static_fields.add("_pytree__node_fields")
        noserialize_fields.add("_pytree__node_fields")

        for field, value in class_vars.items():
            if isinstance(value, dataclasses.Field):
                all_fields[field] = value
                _is_pytree_node = value.metadata.get("pytree_node", True)
                if not _is_pytree_node:
                    if value.metadata.get("ignore", False):
                        ignored_fields.add(field)
                    else:
                        static_fields.add(field)
                elif _is_pytree_node:
                    data_fields.add(field)

                if not value.metadata.get("serialize", True):
                    noserialize_fields.add(field)
                if value.metadata.get("sharded", False):
                    sharded_fields.add(field)

                # default setters
                if value.default is not dataclasses.MISSING:
                    default_setters[field] = partial(identity, value.default)
                elif value.default_factory is not dataclasses.MISSING:
                    default_setters[field] = value.default_factory

            elif isinstance(value, CachedProperty):
                cached_prop_fields.add(field)

            # add setter descriptors
            if hasattr(value, "__set__"):
                setter_descriptors.add(field)

        for field in cached_prop_fields:
            # setattr(cls, _cache_name(field), Uninitialized)
            if class_vars[field].pytree_node:
                data_fields.add(_cache_name(field))
            elif class_vars[field].pytree_ignore:
                ignored_fields.add(_cache_name(field))
            else:
                static_fields.add(_cache_name(field))
        cached_prop_fields = cached_prop_fields.union(
            _inherited_cachedproperty_fields(cls)
        )

        # If no annotations in this class, skip, otherwise we'd process
        # parent's annotations twice
        if "__annotations__" in cls.__dict__:
            # fields that are only type annotations, but do not explicitly declare
            # a struct.field(), feed them forward and treat as standard fields.
            for field, _ in cls.__annotations__.items():
                if field not in all_fields:
                    _value = dataclasses.field()
                    _value.name = field
                    all_fields[field] = _value
                    data_fields.add(field)

                    # If field is in class_vars, it has a default value declared.
                    # this can be overridden by subclasses.
                    if field in class_vars:
                        # has default value
                        value = class_vars[field]
                        default_setters[field] = partial(identity, value)

        # Set mutable to be inherited from parent, or false otherwise
        if mutable is None:
            for parent_class in cls.mro():
                mutable = getattr(parent_class, "_pytree__class_is_mutable", None)
                if mutable is not None:
                    break
        if mutable is None:
            mutable = False

        if mutable and len(cached_prop_fields) != 0:
            raise ValueError("cannot use cached properties with mutable pytrees.")

        if config.netket_sphinx_build:
            for k in static_fields:
                try:
                    delattr(cls, k)
                except AttributeError:
                    pass
            for k in data_fields:
                try:
                    delattr(cls, k)
                except AttributeError:
                    pass

        # new
        init_fields = tuple(
            sorted(data_fields.union(static_fields).union(ignored_fields))
        )
        data_fields = tuple(sorted(data_fields))
        cached_prop_fields = tuple(sorted(cached_prop_fields))
        cached_prop_fields = tuple(_cache_name(f) for f in cached_prop_fields)
        default_setters = MappingProxyType(default_setters)

        static_fields = tuple(sorted(static_fields))
        ignored_fields = tuple(sorted(ignored_fields))
        noserialize_fields = tuple(sorted(noserialize_fields))
        serialize_rename_fields = tuple(
            sorted(compute_serialize_rename_fields(all_fields))
        )
        sharded_fields = tuple(sorted(sharded_fields))

        # init class variables
        cls._pytree__initializing = False
        cls._pytree__class_is_mutable = mutable
        cls._pytree__fields = all_fields
        cls._pytree__static_fields = static_fields
        cls._pytree__ignored_fields = ignored_fields
        cls._pytree__noserialize_fields = noserialize_fields
        cls._pytree__serialize_rename_fields = serialize_rename_fields
        cls._pytree__sharded_fields = sharded_fields
        cls._pytree__setter_descriptors = frozenset(setter_descriptors)
        cls._pytree__default_setters = default_setters

        # new
        cls._pytree__class_dynamic_nodes = dynamic_nodes
        cls._pytree__data_fields = data_fields
        cls._pytree__cachedprop_fields = cached_prop_fields
        cls._pytree__init_fields = init_fields

        jax.tree_util.register_pytree_with_keys(
            cls,
            partial(cls._pytree__flatten, with_key_paths=True),
            cls._pytree__unflatten,
            flatten_func=partial(cls._pytree__flatten, with_key_paths=False),
        )

        serialization.register_serialization_state(
            cls,
            partial(cls._to_flax_state_dict, cls._pytree__noserialize_fields),
            partial(cls._from_flax_state_dict, cls._pytree__noserialize_fields),
        )

    def __pre_init__(self, *args, **kwargs):
        # Default implementation of __pre_init__, used by netket's
        # dataclasses for preinitialisation shuffling of parameters.
        #
        # This is necessary for PyTrees that are subclassed by a dataclass
        # (like a user-implemented sampler using legacy logic).
        #
        # This class takes out all arguments and kw-arguments that are
        # directed to the PyTree from a processing and 'hides' them
        # in a proprietary kwargument for later manipulation.
        #
        # This is necessary so we call the dataclass init only with
        # the arguments that it needs.

        # process keyword arguments
        kwargs_dataclass = {}
        kwargs_pytree = {}
        for k, v in kwargs.items():
            if k in self.__dataclass_fields__.keys():
                kwargs_dataclass[k] = v
            else:
                kwargs_pytree[k] = v

        # process positional args. Identify max positional arguments of the
        # topmost user defined init method
        max_pytree_args = getattr(self, DATACLASS_USER_INIT_N_ARGS, len(args))
        n_args_pytree = min(len(args), max_pytree_args)

        # First n args are for the pytree initialiser (lower) and later
        # positional arguments are for the dataclass initializer
        args_pytree = args[:n_args_pytree]
        args_dataclass = args[n_args_pytree:]

        signature_pytree = (args_pytree, kwargs_pytree)
        kwargs_dataclass["__base_init_args"] = signature_pytree

        return args_dataclass, kwargs_dataclass

    def __post_init__(self):
        pass

    def __process_deserialization_updates__(self, updates):
        """
        Internal function used to modify a posteriori how a
        PyTree is deserialized.

        This is fed a dictionary of {'attribute names':values}
        that will be used to call `self.replace(**updates)`,
        and should return a modified dictionary.

        Can be for example used to support backward compatible
        deserialization.
        """
        return updates

    @classmethod
    def _pytree__flatten(
        cls,
        pytree: "Pytree",
        *,
        with_key_paths: bool,
    ) -> tuple[
        tuple[tp.Any, ...],
        tp.Mapping[str, tp.Any],
    ]:
        all_vars = vars(pytree).copy()
        try:
            static = {k: all_vars.pop(k) for k in pytree._pytree__static_fields}
        except KeyError as e:
            raise KeyError(
                f"Missing static field {e} in PyTree {cls.__name__} while flattening."
            ) from e

        # drop ignored keys
        for ignored_field in pytree._pytree__ignored_fields:
            all_vars.pop(ignored_field, None)
        if with_key_paths:
            node_values = tuple(
                (jax.tree_util.GetAttrKey(field), all_vars.pop(field))
                for field in pytree._pytree__node_fields
            )
        else:
            node_values = tuple(
                all_vars.pop(field) for field in pytree._pytree__node_fields
            )

        if all_vars:
            raise ValueError(
                f"Unexpected fields in {cls.__name__}: {', '.join(all_vars.keys())}."
                "You cannot add new fields to a Pytree after it has been initialized."
            )

        return node_values, MappingProxyType(static)

    @classmethod
    def _pytree__unflatten(
        cls: type[P],
        static_fields: tp.Mapping[str, tp.Any],
        node_values: tuple[tp.Any, ...],
    ) -> P:
        pytree = object.__new__(cls)
        pytree.__dict__.update(zip(static_fields["_pytree__node_fields"], node_values))
        pytree.__dict__.update(static_fields)
        return pytree

    @classmethod
    def _to_flax_state_dict(
        cls, noserialize_field_names: tuple[str, ...], pytree: "Pytree"
    ) -> dict[str, tp.Any]:
        from .pytree_serialization_sharding import to_flax_state_dict_sharding

        state_dict = {
            name: unwrap_jax_prng_keys(
                serialization.to_state_dict(getattr(pytree, name))
            )
            for name in pytree.__dict__
            if name not in noserialize_field_names
        }

        # Handle sharding
        for name in cls._pytree__sharded_fields:
            if name in cls._pytree__noserialize_fields:
                continue
            state_dict[name] = to_flax_state_dict_sharding(state_dict[name])
        # End handle sharding

        # rename fields if we must
        for name, serialize_name in cls._pytree__serialize_rename_fields:
            state_dict[serialize_name] = state_dict[name]
            del state_dict[name]
        # end renaming

        return state_dict

    @classmethod
    def _from_flax_state_dict(
        cls,
        noserialize_field_names: tuple[str, ...],
        pytree: P,
        state: dict[str, tp.Any],
    ) -> P:
        """Restore the state of a data class."""
        state = state.copy()  # copy the state so we can pop the restored fields.
        updates = {}

        # rename fields if we must
        for name, serialize_name in cls._pytree__serialize_rename_fields:
            if serialize_name in state:
                state[name] = state[serialize_name]
                del state[serialize_name]

        # end renaming

        for name in pytree.__dict__:
            if name in noserialize_field_names:
                continue
            if name not in state:
                raise ValueError(
                    f"Missing field {name} in state dict while restoring"
                    f" an instance of {type(pytree).__name__},"
                    f" at path {serialization.current_path()}"
                )
            value = getattr(pytree, name)
            value_state = state.pop(name)
            # updates[name] = serialization.from_state_dict(value, value_state, name=name)

            # handle PRNG arrays: if target is a prng key array we unwrap it
            _pytree_prng_orig_value = dataclasses.MISSING
            if isinstance(value, jax.Array) and jnp.issubdtype(
                value.dtype, jax.dtypes.prng_key
            ):
                _pytree_prng_orig_value = jax.random.key_impl(value)
                value = jax.random.key_data(value)
                # return jax.random.wrap_key_data(source_data, jax.random.key_impl(target_maybe_key))
                # return jax.random.key_data(maybe_key)

            # Handle sharding
            # from netket.utils import mpi
            # import jax
            if name in cls._pytree__sharded_fields:
                updates[name] = (
                    cls._pytree__fields[name]
                    .metadata["sharded"]
                    .deserialization_function(value, value_state, name=name)
                )

            # End handle sharding
            else:
                updates[name] = serialization.from_state_dict(
                    value, value_state, name=name
                )

            # rewrap prng arrays
            if _pytree_prng_orig_value is not dataclasses.MISSING:
                updates[name] = jax.random.wrap_key_data(
                    updates[name], impl=_pytree_prng_orig_value
                )

        if state:
            names = ",".join(state.keys())
            raise ValueError(
                f'Unknown field(s) "{names}" in state dict while'
                f" restoring an instance of {type(pytree).__name__}"
                f" at path {serialization.current_path()}"
            )

        updates = pytree.__process_deserialization_updates__(updates)
        new_pytree = pytree.replace(**updates)
        return new_pytree

    def replace(self: P, **kwargs: tp.Any) -> P:
        """
        Replace the values of the fields of the object with the values of the
        keyword arguments. If the object is a dataclass, `dataclasses.replace`
        will be used. Otherwise, a new object will be created with the same
        type as the original object.
        """
        if dataclasses.is_dataclass(self):
            pytree = dataclasses.replace(self, **kwargs)
        else:
            unknown_keys = set(kwargs) - set(vars(self))
            if unknown_keys:
                raise ValueError(
                    f"Trying to replace unknown fields {unknown_keys} "
                    f"for '{type(self).__name__}'"
                )

            pytree = copy(self)
            pytree.__dict__.update(kwargs)

        # Reset cached properties
        for fname in pytree._pytree__cachedprop_fields:
            setattr(pytree, fname, Uninitialized)

        return pytree

    if not tp.TYPE_CHECKING:

        def __setattr__(self: P, field: str, value: tp.Any):
            if self._pytree__initializing:
                typ = type(self)
                if self._pytree__class_dynamic_nodes:
                    pass
                elif isinstance(getattr(typ, field, None), property):
                    pass
                elif field not in self._pytree__init_fields:
                    raise NetKetPyTreeUndeclaredAttributeAssignmentError(
                        self, field, self._pytree__init_fields
                    )
            else:
                if field in self._pytree__setter_descriptors:
                    pass
                elif field in self._pytree__cachedprop_fields:
                    pass
                elif not hasattr(self, field):
                    raise AttributeError(
                        f"Cannot add new fields to {type(self)} after initialization"
                    )
                elif not self._pytree__class_is_mutable:
                    raise AttributeError(
                        f"{type(self)} is immutable, trying to update field {field}"
                    )

            object.__setattr__(self, field, value)


def _inherited_fields(
    cls: type,
) -> tuple[dict[str, dataclasses.Field], set[str], set[str], set[str]]:
    """
    Returns the set of all fields (static and data) fields from
    base classes.
    """
    fields = {}
    data_fields = set()
    static_fields = set()
    ignored_fields = set()

    for parent_class in cls.mro():
        if parent_class is not cls and parent_class is not Pytree:
            if issubclass(parent_class, Pytree):
                fields.update(parent_class._pytree__fields)
                data_fields.update(parent_class._pytree__data_fields)
                static_fields.update(parent_class._pytree__static_fields)
                ignored_fields.update(parent_class._pytree__ignored_fields)
            elif dataclasses.is_dataclass(parent_class):
                for field in dataclasses.fields(parent_class):
                    fields[field.name] = field

                    _is_pytree_node = field.metadata.get("pytree_node", True)
                    _ignore_node = field.metadata.get("struct_ignore", False)
                    if _is_pytree_node:
                        data_fields.add(field.name)
                    else:
                        if _ignore_node:
                            ignored_fields.add(field.name)
                        else:
                            static_fields.add(field.name)

    return fields, data_fields, static_fields, ignored_fields


def _inherited_noserialize_fields(cls: type) -> set[str]:
    """
    Returns the set of noserialize fields of base classes
    of the input class
    """
    noserialize_fields = set()
    for parent_class in cls.mro():
        if parent_class is not cls and parent_class is not Pytree:
            if issubclass(parent_class, Pytree):
                noserialize_fields.update(parent_class._pytree__noserialize_fields)
            elif dataclasses.is_dataclass(parent_class):
                for field in dataclasses.fields(parent_class):
                    if not field.metadata.get("serialize", True):
                        noserialize_fields.add(field.name)
    return noserialize_fields


def _inherited_sharded_fields(cls: type) -> set[str]:
    """
    Returns the set of noserialize fields of base classes
    of the input class
    """
    sharded_fields = set()
    for parent_class in cls.mro():
        if parent_class is not cls and parent_class is not Pytree:
            if issubclass(parent_class, Pytree):
                sharded_fields.update(parent_class._pytree__sharded_fields)
            elif dataclasses.is_dataclass(parent_class):
                for field in dataclasses.fields(parent_class):
                    if field.metadata.get("sharded", False):
                        sharded_fields.add(field.name)
    return sharded_fields


def _inherited_cachedproperty_fields(cls: type) -> set[str]:
    """
    Returns the set of cached properties of base classes
    of the input class.
    """
    cachedproperty_fields = set()
    for parent_class in cls.mro():
        if parent_class is not cls and parent_class is not Pytree:
            if issubclass(parent_class, Pytree):
                fields = tuple(
                    _raw_cache_name(f) for f in parent_class._pytree__cachedprop_fields
                )
                cachedproperty_fields.update(fields)
            elif dataclasses.is_dataclass(parent_class):
                for field in dataclasses.fields(parent_class):
                    if isinstance(field, CachedProperty):
                        cachedproperty_fields.add(field.name)
    return cachedproperty_fields


def _inherited_defaults_setters(cls: type) -> set[Callable[[], Any]]:
    """
    Returns the set of default setters of base classes
    of the input class.
    """
    default_setters = {}
    for parent_class in cls.mro():
        if parent_class is not cls and parent_class is not Pytree:
            if issubclass(parent_class, Pytree):
                default_setters.update(parent_class._pytree__default_setters)
            elif dataclasses.is_dataclass(parent_class):
                for field in dataclasses.fields(parent_class):
                    if field.default is not dataclasses.MISSING:
                        default_setters[field.name] = lambda: field.default
                    elif field.default_factory is not dataclasses.MISSING:
                        default_setters[field.name] = field.default_factory

    return default_setters


def compute_serialize_rename_fields(all_fields):
    rename = []
    for name, field in all_fields.items():
        new_name = field.metadata.get("serialize_name", None)
        if new_name is not None and new_name != name:
            rename.append((name, new_name))
    return tuple(sorted(rename))


def unwrap_jax_prng_keys(maybe_key):
    if isinstance(maybe_key, jax.Array) and jnp.issubdtype(
        maybe_key.dtype, jax.dtypes.prng_key
    ):
        return jax.random.key_data(maybe_key)
    return maybe_key



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
import warnings

import jax
import jax.numpy as jnp

from netket import config
from netket.errors import UnoptimisedCustomConstraintRandomStateMethodWarning
from netket.hilbert import HomogeneousHilbert
from netket.utils.dispatch import dispatch
from netket.hilbert.constraint import SumConstraint, SumOnPartitionConstraint

from .fock import _random_states_with_constraint_fock


@dispatch
def random_state(hilb: HomogeneousHilbert, key, batches: int, *, dtype=None):
    return random_state(hilb, hilb.constraint, key, batches, dtype=dtype)


@dispatch
@partial(jax.jit, static_argnames=("hilb", "batches", "dtype"))
def random_state(  # noqa: F811
    hilb: HomogeneousHilbert, constraint: None, key, batches: int, *, dtype=None
):
    if dtype is None:
        dtype = hilb._local_states.dtype

    x_ids = jax.random.randint(
        key, shape=(batches, hilb.size), minval=0, maxval=len(hilb._local_states)
    )
    res = hilb.local_indices_to_states(x_ids, dtype=dtype)
    return res


@dispatch
@partial(jax.jit, static_argnames=("hilb", "batches", "dtype"))
def random_state(  # noqa: F811
    hilb: HomogeneousHilbert,
    constraint: SumConstraint,
    key,
    batches: int,
    *,
    dtype=None,
):
    local_states = hilb._local_states
    if dtype is None:
        dtype = hilb._local_states.dtype

    # Convert total constraint to Fock-like total number of excitations
    n_excitations = (
        constraint.sum_value - (local_states.start * hilb.size)
    ) // local_states.step

    samples_indx = _random_states_with_constraint_fock(
        n_excitations, hilb.shape, key, (batches,), dtype
    )
    return hilb.local_indices_to_states(samples_indx, dtype=dtype)


@dispatch
@partial(jax.jit, static_argnames=("hilb", "batches", "dtype"))
def random_state(  # noqa: F811
    hilb: HomogeneousHilbert,
    constraint: SumOnPartitionConstraint,
    key,
    batches: int,
    *,
    dtype=None,
):
    if dtype is None:
        dtype = hilb._local_states.dtype

    # Convert total constraint to Fock-like total number of excitations
    local_states = hilb._local_states
    n_excitations = [
        (sum_val - (local_states.start * hilb.size)) // local_states.step
        for sum_val in constraint.sum_values
    ]

    n_subspaces = len(n_excitations)
    keys = jax.random.split(key, len(n_excitations))
    shape = hilb.shape
    sizes = (0,) + constraint.sizes

    # Generate a valid configuration for every sub-partition
    vs = [
        _random_states_with_constraint_fock(
            n_excitations[i],
            shape[sum(sizes[: i + 1]) : sum(sizes[: i + 2])],
            keys[i],
            (batches,),
            dtype,
        )
        for i in range(n_subspaces)
    ]
    # then concatenate
    samples_indx = jnp.concatenate(vs, axis=-1)

    return hilb.local_indices_to_states(samples_indx, dtype=dtype)


@dispatch
@partial(jax.jit, static_argnames=("hilb", "batches", "dtype"))
def random_state(  # noqa: F811
    hilb: HomogeneousHilbert, constraint, key, batches: int, *, dtype=None
):
    if config.netket_random_state_fallback_warning:
        warnings.warn(
            UnoptimisedCustomConstraintRandomStateMethodWarning(hilb, constraint)
        )

    keys = jax.random.split(key, batches + 1)
    states = random_state(hilb, None, keys[0], batches, dtype=dtype)

    def _loop_until_ok(state, key):
        def __body(args):
            state, _key = args
            _key, subkey = jax.random.split(_key)
            new_state = random_state(hilb, None, subkey, 1, dtype=dtype)[0]
            return (new_state, _key)

        def __cond(args):
            state, _ = args
            return jnp.logical_not(constraint(state))

        return jax.lax.while_loop(__cond, __body, (state, key))[0]

    return jax.vmap(_loop_until_ok, in_axes=(0, 0))(states, keys[1:])


@dispatch
def flip_state_scalar(hilb: HomogeneousHilbert, key, σ, idx):  # noqa: F811
    local_dimension = len(hilb._local_states)

    if local_dimension < 2:
        return σ, σ[idx]

    # Get site to flip, convert that individual site to indices
    σi_old = σ[idx]
    xi_old = hilb.states_to_local_indices(σi_old)

    if local_dimension == 2:
        # hardcode for 2-dim, where there is no randomness
        xi_new = jnp.where(xi_old == 1, 0, 1).astype(xi_old.dtype)
    else:
        # compute flipped index
        r = jax.random.uniform(key)
        xi_new = jax.numpy.floor(r * (local_dimension - 1))
        xi_new = xi_new + (xi_new >= xi_old)
        xi_new = xi_new.astype(xi_old.dtype)

    # return
    σ_new = σ.at[idx].set(hilb.local_indices_to_states(xi_new, dtype=σ.dtype))

    return σ_new, σi_old



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from . import constraint
from . import index

from .abstract_hilbert import AbstractHilbert
from .discrete_hilbert import DiscreteHilbert
from .homogeneous import HomogeneousHilbert


from .doubled_hilbert import DoubledHilbert
from .spin import Spin
from .fock import Fock
from .qubit import Qubit
from .spin_orbital_fermions import SpinOrbitalFermions

from .tensor_hilbert import TensorHilbert
from . import tensor_hilbert_discrete

from . import random

from netket.utils import _hide_submodules

_hide_submodules(__name__)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
import jax.numpy as jnp

from netket.utils import random_seed, mpi, config
from netket.utils.mpi import MPI_jax_comm
from netket.utils.types import PRNGKeyT, SeedT


def PRNGKey(seed: SeedT | None = None, *, root: int = 0, comm=MPI_jax_comm) -> PRNGKeyT:
    """
    Initialises a PRNGKey using an optional starting seed.

    If using sharding, the returned key will be replicated while if using MPI
    the key of the master rank will be broadcasted to every process.

    Args:
        seed: An optional integer value to use as seed
        root: the master rank, used when running under MPI (defaults to 0)
        comm: The MPI communicator to use for broadcasting, if necessary

    Returns:
        A sharded/broadcasted :func:`jax.random.PRNGKey`.

    """
    if seed is None:
        seed = random_seed()

    if isinstance(seed, int):
        # We can't sync the PRNGKey, so we can only sinc integer seeds
        # see https://github.com/google/jax/pull/16511
        if config.netket_experimental_sharding and jax.process_count() > 1:  # type: ignore[attr-defined]
            # TODO: use stable jax function
            from jax.experimental import multihost_utils

            seed = int(
                multihost_utils.broadcast_one_to_all(
                    seed, is_source=jax.process_index() == root
                ).item()
            )

        key = jax.random.PRNGKey(seed)
    else:
        key = seed

    if config.netket_experimental_sharding:
        key = jax.lax.with_sharding_constraint(
            key, jax.sharding.PositionalSharding(jax.devices()).replicate()
        )
    else:  # type: ignore[attr-defined]
        key = _bcast_key(key, root=root, comm=comm)
    return key


class PRNGSeq:
    """
    A sequence of PRNG keys generated based on an initial key.
    """

    def __init__(self, base_key: SeedT | None = None):
        if base_key is None:
            base_key = PRNGKey()
        elif isinstance(base_key, int):
            base_key = PRNGKey(base_key)
        self._current: PRNGKeyT = base_key

    def __iter__(self):
        return self

    def __next__(self):
        self._current = jax.random.split(self._current, num=1)[0]
        return self._current

    def next(self):
        return self.__next__()

    def take(self, num: int):
        """
        Returns an array of `num` PRNG keys and advances the iterator accordingly.
        """
        keys = jax.random.split(self._current, num=num + 1)
        self._current = keys[-1]
        return keys[:-1]


def mpi_split(key, *, root=0, comm=MPI_jax_comm) -> PRNGKeyT:
    """
    Split a key across MPI nodes in the communicator.
    Only the input key on the root process matters.

    Arguments:
        key: The key to split. Only considered the one on the root process.
        root: (default=0) The root rank from which to take the input key.
        comm: (default=MPI.COMM_WORLD) The MPI communicator.

    Returns:
        A PRNGKey depending on rank number and key.
    """

    # Maybe add error/warning if in_key is not the same
    # on all MPI nodes?
    keys = jax.random.split(key, mpi.n_nodes)

    keys = _bcast_key(keys, root=root, comm=comm)

    return keys[mpi.rank]


def _bcast_key(key, root=0, comm=MPI_jax_comm) -> PRNGKeyT:
    """
    Utility function equivalent to calling `mpi_bcast_jax` on a jax key,
    but working around some sharding bug when not using sharding, arising
    from MPI.
    """
    is_new_style_key = jnp.issubdtype(key.dtype, jax.dtypes.prng_key)

    if is_new_style_key:
        _impl = jax.random.key_impl(key)
        key = jax.random.key_data(key)

    key = jax.tree_util.tree_map(
        lambda k: mpi.mpi_bcast_jax(k, root=root, comm=comm)[0], key
    )

    if is_new_style_key:
        key = jax.random.wrap_key_data(key, impl=_impl)  # type: ignore[arg-type]

    return key


def batch_choice(key, a, p):
    """
    Batched version of `jax.random.choice`.

    Attributes:
      key: a PRNGKey used as the random key.
      a: 1D array. Random samples are generated from its elements.
      p: 2D array of shape `(batch_size, a.size)`. Each slice `p[i, :]` is
        the probabilities associated with entries in `a` to generate a sample
        at the index `i` of the output. Can be unnormalized.

    Returns:
      The generated samples as an 1D array of shape `(batch_size,)`.
    """
    p_cumsum = p.cumsum(axis=1)
    r = p_cumsum[:, -1:] * jax.random.uniform(key, shape=(p.shape[0], 1))
    indices = (r > p_cumsum).sum(axis=1)
    out = a[indices]
    return out



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import dataclasses
import abc

PyTree = Any


@dataclasses.dataclass(frozen=True)
class ModuleFramework(abc.ABC):

    @property
    def model_contains_parameters(self) -> bool:
        """
        Returns True if the model contains the parameters in the model itself, False
        if the parameters are stored separately.
        """
        return False

    @staticmethod
    @abc.abstractmethod
    def is_loaded() -> bool:
        """
        Returns True if this module framework has already been loaded by the user.
        """

    @staticmethod
    @abc.abstractmethod
    def is_my_module(module: Any) -> bool:
        """
        Returns True if the given module is from this framework, False otherwise.

        Args:
            module: a module from an unknown framework.
        """

    @staticmethod
    @abc.abstractmethod
    def wrap(module: Any) -> tuple[PyTree | None, Any]:
        """
        Wraps the given module in a way that it behaves like a flax module, possibly
        returning the parameters as well.

        For flax-like modules, this should return None and the module itself. For
        modules that store the parameters in the module itself, it should return the
        parameters and a static object that can be used to apply the module.

        Args:
            A module from the framework corresponding to this class.

        Returns:
            A tuple with the parameters, if any, and the static module.
        """
        raise NotImplementedError

    @staticmethod
    def unwrap(module: Any, maybe_variables: PyTree | None) -> Any:
        """
        Undoes the wrapping done by `wrap`, restoring the original module.

        For flax-like modules, this should do nothing. For modules that store
        the parameters in the module itself, it should unwrap the module and
        restore the parameters.

        Args:
            module: the module to unwrap
            maybe_variables: the variables obtained from the wrapping, if any.

        Returns:
            The original module from this framework.
        """
        return module


registered_frameworks = []


def framework(clz):
    """
    Registers a framework and it's wrapper methods to make it
    behave like a flax framework.
    """
    clz = dataclasses.dataclass(frozen=True)(clz)
    registered_frameworks.append(clz)
    return clz


@dataclasses.dataclass(frozen=True)
class UnknownFramework(ModuleFramework):
    name: str = "Unknown"

    @staticmethod
    def is_loaded() -> bool:
        return True

    @staticmethod
    def is_my_module(module) -> bool:
        return False

    @staticmethod
    def wrap(module) -> tuple:
        return None, module

    @staticmethod
    def unwrap(wrapped_module, wrapped_variables):
        return wrapped_module


def identify_framework(module):
    for _framework in registered_frameworks:
        if _framework.is_loaded() and _framework.is_my_module(module):
            return _framework

    return UnknownFramework


def maybe_wrap_module(module) -> tuple:
    """
    Passing a module from an unknown framework (might be user defined module, a jax
    module, flax or haiku or anything else really), attempt to identify what is the
    package/framework it comes from, and if so it correctly wraps it in order to
    make it behave like a flax module (our default).

    Also returns a function used to unpack the parameters once we are done.
    """
    framewrk = identify_framework(module)

    maybe_module_variables, static_module = framewrk.wrap(module)

    return maybe_module_variables, static_module



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable

from jax import numpy as jnp
from jax.nn.initializers import zeros

from netket.graph import AbstractGraph
from netket.models.autoreg import _get_feature_list
from netket.models.fast_autoreg import FastARNNSequential
from netket.utils import HashableArray
from netket.utils.types import Array, DType, NNInitFunc

from netket.experimental.models.rnn import RNN
from netket.experimental.nn.rnn import (
    FastRNNLayer,
    GRU1DCell,
    LSTMCell,
    default_kernel_init,
    ensure_prev_neighbors,
)


class FastRNN(FastARNNSequential):
    """
    Base class for recurrent neural networks with fast sampling.

    The fast autoregressive sampling is described in `Ramachandran et. {\\it al} <https://arxiv.org/abs/1704.06001>`_.
    To generate one sample using an autoregressive network, we need to evaluate the network `N` times, where `N` is
    the number of input sites. But actually we only change one input site each time, and not all intermediate results
    depend on the changed input because of the autoregressive property, so we can cache unchanged intermediate results
    and avoid repeated computation.

    This optimization is particularly useful for RNN where each output site of a layer only depends on a small number of
    input sites. In the slow RNN, we need to run `N` RNN steps in each layer during each AR sampling step. While in the
    fast RNN, we cache the relevant hidden memories in each layer from the previous AR sampling step, and only run one
    RNN step to update from the changed input.

    See :class:`netket.experimental.models.RNN` for explanation of the arguments related to
    the autoregressive order.
    """

    layers: int
    """number of layers."""
    features: Iterable[int] | int
    """output feature density in each layer. If a single number is given,
    all layers except the last one will have the same number of features."""
    reorder_idx: HashableArray | None = None
    """indices to transform the inputs from unordered to ordered.
    See :meth:`netket.models.AbstractARNN.reorder` for details."""
    inv_reorder_idx: HashableArray | None = None
    """indices to transform the inputs from ordered to unordered.
    See :meth:`netket.models.AbstractARNN.reorder` for details."""
    prev_neighbors: HashableArray | None = None
    """previous neighbors of each site."""
    graph: AbstractGraph | None = None
    """graph of the physical system."""
    param_dtype: DType = jnp.float64
    """the dtype of the computation (default: float64)."""
    kernel_init: NNInitFunc = default_kernel_init
    """initializer for the weights."""
    bias_init: NNInitFunc = zeros
    """initializer for the biases."""
    machine_pow: int = 2
    """exponent to normalize the outputs of `__call__`."""

    def __post_init__(self):
        reorder_idx, inv_reorder_idx, prev_neighbors = ensure_prev_neighbors(
            reorder_idx=self.reorder_idx,
            inv_reorder_idx=self.inv_reorder_idx,
            prev_neighbors=self.prev_neighbors,
            graph=self.graph,
        )

        self.reorder_idx = reorder_idx
        self.inv_reorder_idx = inv_reorder_idx
        self.prev_neighbors = prev_neighbors
        super().__post_init__()

    def reorder(self, inputs: Array, axis: int = 0) -> Array:
        return RNN.reorder(self, inputs, axis)

    def inverse_reorder(self, inputs: Array, axis: int = 0) -> Array:
        return RNN.inverse_reorder(self, inputs, axis)

    def _take_prev_site(self, inputs: Array, index: int) -> Array:
        if self.reorder_idx is None:
            k = index
            prev_index = k - 1
        else:
            k = jnp.asarray(self.inv_reorder_idx)[index]
            prev_index = jnp.asarray(self.reorder_idx)[k - 1]

        inputs_i = inputs[:, prev_index, :]
        inputs_i = jnp.where(k == 0, 0, inputs_i)
        return inputs_i


class FastLSTMNet(FastRNN):
    """
    Long short-term memory network with fast sampling.

    See :class:`netket.models.FastARNNSequential` for a brief explanation of fast autoregressive sampling.
    """

    def setup(self):
        features = _get_feature_list(self)
        self._layers = [
            FastRNNLayer(
                cell=LSTMCell(
                    features=features[i],
                    param_dtype=self.param_dtype,
                    kernel_init=self.kernel_init,
                    bias_init=self.bias_init,
                ),
                size=self.hilbert.size,
                exclusive=(i == 0),
                reorder_idx=self.reorder_idx,
                inv_reorder_idx=self.inv_reorder_idx,
                prev_neighbors=self.prev_neighbors,
            )
            for i in range(self.layers)
        ]


class FastGRUNet1D(FastRNN):
    """
    Gated recurrent unit network with fast sampling. Only supports one previous neighbor at each site.

    See :class:`netket.models.FastARNNSequential` for a brief explanation of fast autoregressive sampling.
    """

    def setup(self):
        features = _get_feature_list(self)
        self._layers = [
            FastRNNLayer(
                cell=GRU1DCell(
                    features=features[i],
                    param_dtype=self.param_dtype,
                    kernel_init=self.kernel_init,
                    bias_init=self.bias_init,
                ),
                size=self.hilbert.size,
                exclusive=(i == 0),
                reorder_idx=self.reorder_idx,
                inv_reorder_idx=self.inv_reorder_idx,
                prev_neighbors=self.prev_neighbors,
            )
            for i in range(self.layers)
        ]



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, TYPE_CHECKING
from functools import partial

import jax
from plum import Callable

import netket.jax as nkjax
from netket.optimizer.linear_operator import LinearOperator

from .qgt_jacobian import QGTJacobianDense, QGTJacobianPyTree
from .qgt_onthefly import QGTOnTheFly

from .. import solver as nk_solver_module

if TYPE_CHECKING:
    from netket.vqs import VariationalState

QGTConstructor = Callable[["VariationalState"], LinearOperator]

solvers: list[Callable] = []

for solver in dir(nk_solver_module):
    # only add solvers, not random
    # useless things
    if solver[:2] == "__":
        continue
    else:
        solvers.append(getattr(nk_solver_module, solver))


def _is_dense_solver(solver: Any) -> bool:
    """
    Returns true if the solver is one of our known dense solvers
    """
    if isinstance(solver, partial):
        solver = solver.func

    if solver in solvers:
        return True

    return False


def default_qgt_matrix(
    variational_state, solver: Any = False, **kwargs
) -> QGTConstructor:
    """
    Determines default metric tensor depending on variational_state and solver
    """
    from netket.vqs import FullSumState

    if isinstance(variational_state, FullSumState):
        return partial(QGTJacobianPyTree, **kwargs)

    n_param_leaves = len(jax.tree_util.tree_leaves(variational_state.parameters))
    n_params = variational_state.n_parameters

    # those require dense matrix that is known to be faster for this qgt
    if _is_dense_solver(solver):
        return partial(QGTJacobianDense, **kwargs)

    # TODO: Remove this once all QGT support diag_scale.
    has_diag_rescale = kwargs.get("diag_scale") is not None

    # arbitrary heuristic: if the network's parameters has many leaves
    # (an rbm has 3) then JacobianDense might be faster
    # the numbers chosen below are rather arbitrary and should be tuned.
    if (n_param_leaves > 6 and n_params > 800) or has_diag_rescale:
        if nkjax.tree_ishomogeneous(variational_state.parameters):
            return partial(QGTJacobianDense, **kwargs)
        else:
            return partial(QGTJacobianPyTree, **kwargs)
    else:
        return partial(QGTOnTheFly, **kwargs)  # type: ignore


class QGTAuto:
    """
    Automatically select the 'best' Quantum Geometric Tensor
    computing format according to some rather untested heuristic.

    Args:
        variational_state: The variational State
        kwargs: are passed on to the QGT constructor.
    """

    _last_vstate = None
    """Cached last variational state to skip logic to decide what type of
    QGT to chose.
    """

    _last_matrix: QGTConstructor | None = None
    """
    Cached last QGT. Used when vstate == _last_vstate
    """

    _kwargs: dict = {}
    """
    Kwargs passed at construction. Used when constructing a QGT.
    """

    _solver: Callable | None

    def __init__(self, solver: Callable | None = None, **kwargs):
        self._solver = solver

        self._kwargs = kwargs

    def __call__(
        self, variational_state: "VariationalState", *args, **kwargs
    ) -> LinearOperator:
        if self._last_vstate != variational_state:
            self._last_vstate = variational_state

            self._last_matrix = default_qgt_matrix(
                variational_state, solver=self._solver, **self._kwargs, **kwargs
            )

        return self._last_matrix(variational_state, *args, **kwargs)  # type: ignore

    def __repr__(self):
        return "QGTAuto()"



from dataclasses import dataclass
from typing import Sequence

import jax.numpy as jnp
import numpy as np
from jax.tree_util import register_pytree_node_class


@register_pytree_node_class
@dataclass(frozen=True)
class Cell:
    """A finite region in continuous space."""

    extent: tuple[float, ...]
    pbc: tuple[bool, ...]

    def __init__(
        self,
        d: int,
        L: float | Sequence[float] = 1.0,
        pbc: bool | Sequence[bool] = True,
    ) -> None:
        if d <= 0:
            raise ValueError("d must be positive")

        if hasattr(L, "__len__"):
            if len(L) != d:
                raise ValueError("Length of L must match d")
            extent = tuple(float(x) for x in L)
        else:
            extent = (float(L),) * d

        if isinstance(pbc, bool):
            pbc_tuple = (pbc,) * d
        else:
            if len(pbc) != d:
                raise ValueError("Length of pbc must match d")
            pbc_tuple = tuple(bool(x) for x in pbc)

        for e, b in zip(extent, pbc_tuple):
            if np.isinf(e) and b:
                raise ValueError(
                    "Cannot combine periodic boundary conditions with infinite extent"
                )

        object.__setattr__(self, "extent", extent)
        object.__setattr__(self, "pbc", pbc_tuple)

    def tree_flatten(self):
        data = (jnp.asarray(self.extent), jnp.asarray(self.pbc))
        return data, None

    @classmethod
    def tree_unflatten(cls, aux, data):
        extent, pbc = data
        return cls(
            d=len(extent),
            L=tuple(float(x) for x in extent),
            pbc=tuple(bool(x) for x in pbc),
        )

    @property
    def dimension(self) -> int:
        return len(self.extent)

    def distance(
        self, r1: Sequence[float] | jnp.ndarray, r2: Sequence[float] | jnp.ndarray
    ) -> jnp.ndarray:
        """Euclidean distance between ``r1`` and ``r2`` considering boundary conditions."""

        r1 = jnp.asarray(r1)
        r2 = jnp.asarray(r2)
        diff = r1 - r2
        ext = jnp.asarray(self.extent)
        pbc = jnp.asarray(self.pbc)

        shift = jnp.where(pbc, jnp.round(diff / ext) * ext, 0.0)
        diff = diff - shift
        return jnp.linalg.norm(diff, axis=-1)


@register_pytree_node_class
@dataclass(frozen=True)
class FreeSpace(Cell):
    """Infinite space without periodic boundary conditions."""

    def __init__(self, d: int) -> None:
        super().__init__(d=d, L=(np.inf,) * d, pbc=False)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_graph import AbstractGraph
from .graph import Graph, Edgeless, DoubledGraph, disjoint_union
from .lattice import Lattice
from .common_lattices import (
    Grid,
    Hypercube,
    Cube,
    Square,
    Chain,
    BCC,
    FCC,
    Diamond,
    Pyrochlore,
    Triangular,
    Honeycomb,
    Kagome,
    KitaevHoneycomb,
)

from netket.utils import _hide_submodules, _auto_export

_auto_export(__name__)
_hide_submodules(__name__)



import jax
import jax.numpy as jnp

from jax import Array
from flax import struct

from ._sort import searchsorted


@struct.dataclass
class COOArray:
    """
    simple sparse coo array written in jax (similar to sparse.COO)
    supports arbitrary number of dimensions
    so far only indexing and conversion to/from dense is implemented

    Args:
        _coords: array containing the indices of the nonzero elements
        data: array containing the values of the nonzero elements
        shape: shape of the array
        fill_value: default value of the nonzero elements (0)
    """

    _coords: (
        Array  # needs to be unique and sorted, these are transposed w.r.t sparse.COO
    )
    data: Array
    shape: tuple = struct.field(pytree_node=False, default=())
    fill_value: Array = 0

    def __getitem__(self, *idx):
        assert self.coords.ndim == 2
        assert self.data.ndim == 1
        assert self._coords.shape[0] == self.data.shape[0]
        assert self._coords.shape[1] == len(self.shape)
        # TODO support more indexing modes
        if len(idx) == 1 and isinstance(idx[0], tuple):
            idx = jnp.array(idx[0]).T
            if not idx.shape[-1] == self.ndim:
                raise NotImplementedError
        elif len(idx) == self.ndim:
            idx = jnp.array(idx[0])
            if not idx.shape[-1] == self.ndim:
                raise NotImplementedError
        else:
            raise NotImplementedError
        # TODO implement searchsorted with tuples?
        k = searchsorted(self._coords, idx)
        mask = (self._coords[k] == idx).all(axis=-1)
        res = self.data[k]
        return jax.lax.select(mask, res, jnp.full_like(res, self.fill_value))

    @property
    def coords(self) -> Array:
        # for consistency with sparse
        # TODO implement transposed version?
        return self._coords.T

    @property
    def dtype(self):
        return self.data.dtype

    @property
    def ndim(self) -> int:
        return len(self.shape)

    @property
    def nnz(self) -> int:
        return self.data.size

    def todense(self) -> Array:
        A = jnp.full(self.shape, self.fill_value, self.dtype)
        return A.at[tuple(self.coords)].set(self.data)

    @classmethod
    def fromdense(cls, A: Array) -> "COOArray":
        # assume that nonzero returns a sorted array
        i = jnp.nonzero(A)
        return cls(jnp.array(i).T, A[i], A.shape)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from fractions import Fraction
from collections.abc import Sequence


import numpy as np

from netket.utils import StaticRange

from .homogeneous import HomogeneousHilbert
from .constraint import DiscreteHilbertConstraint, SumConstraint


def _check_total_sz(total_sz, S, size):
    if total_sz is None:
        return

    local_size = 2 * S + 1

    m = round(2 * total_sz)
    if np.abs(m) > size * (2 * S):
        raise ValueError(
            "Cannot fix the total magnetization: 2|M| cannot " "exceed Nspins."
        )

    # If half-integer spins (1/2, 3/2)
    if local_size % 2 == 0:
        # Check that the total magnetization is odd if odd spins or even if even
        # number of spins
        if (size + m) % 2 != 0:
            raise ValueError(
                "Cannot fix the total magnetization: Nspins + 2*totalSz must be even."
            )
    # else if full-integer (S=1,2)
    else:
        if m % 2 != 0:
            raise ValueError(
                "Cannot fix the total magnetization to a half-integer number"
            )


class Spin(HomogeneousHilbert):
    r"""Hilbert space obtained as tensor product of local spin states.

    The correspondance between basis elements of the hilbert space and the configurations
    that are then fed to variational states is as follows:

    +----------+----------------------------------+-----------------------------------+
    | State    | Old Behavior                     | New Behavior                      |
    |          | :code:`inverted_ordering=True`   | :code:`inverted_ordering=False`   |
    +==========+==================================+===================================+
    | ↑ ↑ ↑    | -1 -1 -1                         | +1 +1 +1                          |
    +----------+----------------------------------+-----------------------------------+
    | ↑ ↑ ↓    | -1 -1 +1                         | +1 +1 -1                          |
    +----------+----------------------------------+-----------------------------------+
    | ↑ ↓ ↑    | -1 +1 -1                         | +1 -1 +1                          |
    +----------+----------------------------------+-----------------------------------+
    | ↑ ↓ ↓    | -1 +1 +1                         | +1 -1 -1                          |
    +----------+----------------------------------+-----------------------------------+
    | ↓ ↑ ↑    | +1 -1 -1                         | -1 +1 +1                          |
    +----------+----------------------------------+-----------------------------------+
    | ↓ ↑ ↓    | +1 -1 +1                         | -1 +1 -1                          |
    +----------+----------------------------------+-----------------------------------+
    | ↓ ↓ ↑    | +1 +1 -1                         | -1 -1 +1                          |
    +----------+----------------------------------+-----------------------------------+
    | ↓ ↓ ↓    | +1 +1 +1                         | -1 -1 -1                          |
    +----------+----------------------------------+-----------------------------------+

    The old behaviour is the default behaviour of NetKet 3.14 and before, while the new
    behaviour will become the default starting 1st january 2025.
    For that reason, in the transition period, we will print warnings asking to explicitly
    specify which ordering you want

    .. warning::

        The ordering of the Spin Hilbert space basis has historically always been
        such that `-1=↑, 1=↓`, but it will be changed 1st january 2025 to
        be such that `1=↑, -1=↓`.

        The change will break:
            - code that relies on the assumption that -1=↑;
            - all saves because the inputs to the network will change;
            - custom operators that rely on the basis being ordered;

        To avoid distruption, NetKet will support **both** conventions in the (near)
        future. You can specify the ordering you need with :code:`inverted_ordering = True`
        (historical ordering) or :code:`inverted_ordering=False` (future default behaviour).

        If you do not specify this flag, a future version of NetKet might break your
        serialized weights or other logic, so we strongly reccomend that you either
        limit yourself to NetKet 3.14, or that you specify :code:`inverted_ordering`
        explicitly.

    """

    def __init__(
        self,
        s: float,
        N: int = 1,
        *,
        total_sz: float | None = None,
        constraint: DiscreteHilbertConstraint | None = None,
        inverted_ordering: bool = False,
    ):
        r"""Hilbert space obtained as tensor product of local spin states.

        .. note::

            Since NetKet 3.16 (January 2025) the default ordering of the Spin Hilbert space
            basis has changed. The new default is such that `1=↑, -1=↓`. This change can
            be controlled by the `inverted_ordering` flag. If you do not specify this flag,
            you will get the new behaviour.

            To ensure that the old behaviour is maintained, you should specify
            `inverted_ordering=True`. If you want to opt into the new default
            you should specify `inverted_ordering=False`.

        Args:
            s: Spin at each site. Must be integer or half-integer.
            N: Number of sites (default=1)
            total_sz: If given, constrains the total spin of system to a particular
                value.
            constraint: A custom constraint on the allowed configurations. This argument
                cannot be specified at the same time as :code:`total_sz`. The constraint
                must be a subclass of :class:`~netket.hilbert.DiscreteHilbertConstraint`.
            inverted_ordering: Flag to specify the ordering of the Local basis. Historically
                NetKet has always used the convention `-1=↑, 1=↓` (corresponding to
                :code:`inverted_ordering=True`, but we will change it to `1=↑, -1=↓` (
                :code:`inverted_ordering=False`).
                The default as of September 2024 (NetKet 3.14) is :code:`inverted_ordering=True`, but
                we will change it in the near future.
                The change will (i) break code that relies on the assumption that -1=↑, and
                (ii) will break all saves because the inputs to the network will change.


        Examples:
           Simple spin hilbert space.

           >>> import netket as nk
           >>> hi = nk.hilbert.Spin(s=1/2, N=4)
           >>> print(hi.size)
           4
        """
        local_size = round(2 * s + 1)
        assert int(2 * s + 1) == local_size

        # TODO: Remove in NetKet 3.17
        if inverted_ordering is None:
            inverted_ordering = False

        if not inverted_ordering:
            # Reasonable, new ordering where  1=↑ -1=↓
            local_states = StaticRange(
                local_size - 1,  # type: ignore[arg-type]
                -2,  # type: ignore[arg-type]
                local_size,
            )
        else:
            # Old ordering where -1=↑ 1=↓
            local_states = StaticRange(
                1 - local_size,  # type: ignore[arg-type]
                2,  # type: ignore[arg-type]
                local_size,
            )

        _check_total_sz(total_sz, s, N)
        if total_sz is not None:
            if constraint is not None:
                raise ValueError(
                    "Cannot specify at the same time a total magnetization "
                    "constraint and a `custom_constraint."
                )
            constraint = SumConstraint(round(2 * total_sz))

        self._total_sz = total_sz
        self._s = s
        self._inverted_ordering = inverted_ordering

        super().__init__(local_states, N, constraint=constraint)

    def __pow__(self, n):
        if not self.constrained:
            return Spin(self._s, self.size * n)

        return NotImplemented

    def _mul_sametype_(self, other):
        assert type(self) == type(other)
        if self._s == other._s:
            if not self.constrained and not other.constrained:
                return Spin(s=self._s, N=self.size + other.size)

        return NotImplemented

    def ptrace(self, sites: int | Sequence[int]) -> Optional["Spin"]:
        if isinstance(sites, int):
            sites = [sites]

        for site in sites:
            if site < 0 or site >= self.size:
                raise ValueError(
                    f"Site {site} not in this hilbert space of site {self.size}"
                )

        if self.constrained:
            raise TypeError("Cannot take the partial trace with a constraint.")

        Nsites = len(sites)

        if self.size - Nsites == 0:
            return None
        else:
            return Spin(s=self._s, N=self.size - Nsites)

    def __repr__(self):
        if self._total_sz is not None:
            constraint = f", total_sz={self._total_sz}"
        elif self.constrained:
            constraint = f", {self._constraint}"
        else:
            constraint = ""
        ordering = "inverted" if self._inverted_ordering else "new"
        return f"Spin(s={Fraction(self._s)}, N={self.size}, ordering={ordering}{constraint})"

    @property
    def _attrs(self):
        return (self.size, self._s, self._inverted_ordering, self.constraint)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from typing import Any, TYPE_CHECKING

import jax

from netket.utils import mpi

if TYPE_CHECKING:
    from netket.vqs import VariationalState


class AbstractLog(abc.ABC):
    """
    Abstract base class detailing the interface that loggers must
    implement in order to work with netket drivers.

    It can be passed with keyword argument `out` to Monte Carlo drivers in order
    to serialize the output data of the simulation.

    .. note::

        When using Loggers in a multi-process (MPI or Sharding) context, some care
        must be taken to ensure that they work correctly.

        The design philosophy adopted by NetKet follows the Jax/Orbax convention
        that the same code should be executd on all processes. Therefore, loggers
        should expect to be executed and called from ALL processes, and it is
        their responsability to only perform expensive I/O operations on the root
        rank.

        They can verify if they are running on the root rank by calling
        `self._is_master_process`.

        Have a look at :class:`netket.logging.RuntimeLog` or
        :class:`netket.logging.TensorBoardLog` for a good example.

    """

    @abc.abstractmethod
    def __call__(
        self,
        step: int,
        item: dict[str, Any],
        variational_state: "VariationalState | None" = None,
    ):
        """
        Logs at a given integer step a dictionary of data, optionally
        specifying a variational state to encode additional data.

        Args:
            step: monotonically increasing integer representing the row in the
                database corresponding to this log entry;
            item: Any dictionary of data to be logged;
            variational_state: optional variational state from which additional data
                might be extracted.
        """

    @abc.abstractmethod
    def flush(self, variational_state: "VariationalState | None" = None):
        """
        Flushes the data that is stored internally to disk/network.

        Args:
            variational_state: optional variational state from which additional data
                might be extracted.

        """

    @property
    def _is_master_process(self) -> bool:
        """
        Returns whether this logger is the root logger in a distributed setting.
        """
        return mpi.rank == 0 and jax.process_index() == 0



# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from typing import Generic, TypeVar

from netket.hilbert import AbstractHilbert


HilbertType = TypeVar("HilbertType", bound=AbstractHilbert)


class AbstractObservable(ABC, Generic[HilbertType]):
    """Abstract class for quantum Observables.

    An observable is a general object that defines a quantity that
    can be comptued starting from a variational state. Observables
    should be computed using the method `expect` of the variational
    states, and derivatives of this expectation value can sometimes
    also be computed using the method `expect_and_grad`.

    All operators are Observables, but some observables are not
    operators (for example, the entanglement entropy observable
    does is not an operator).

    This class determines the basic methods that an observable
    must implement to work correctly with NetKet.
    """

    _hilbert: HilbertType
    r"""The hilbert space associated to this observable."""

    def __init__(self, hilbert: HilbertType):
        self._hilbert = hilbert

    @property
    def hilbert(self) -> HilbertType:
        r"""The hilbert space associated to this observable."""
        return self._hilbert

    def __repr__(self):
        return f"{type(self).__name__}(hilbert={self.hilbert})"
